from .tensor import Tensor, swapaxes


def sqrt(x: Tensor):
    '''平方根函数'''
    return x**0.5


def square(x: Tensor):
    '''平方函数'''
    return x * x


def vsplit(x: Tensor, indices_or_sections: int | tuple) -> list[Tensor]:
    if not isinstance(x, Tensor):
        x = Tensor(x)

    try:
        len(indices_or_sections)
    except TypeError:
        sections = indices_or_sections
        N = x.shape[0]
        assert N % sections == 0, 'array split does not result in an equal division'

    Ntotal = x.shape[0]
    try:
        # handle array case.
        Nsections = len(indices_or_sections) + 1
        div_points = [0] + list(indices_or_sections) + [Ntotal]
    except TypeError:
        # indices_or_sections is a scalar, not an array.
        Nsections = int(indices_or_sections)
        if Nsections <= 0:
            raise ValueError(
                'number sections must be larger than 0.') from None
        Neach_section, extras = divmod(Ntotal, Nsections)
        section_sizes = ([0] + extras * [Neach_section + 1] +
                         (Nsections - extras) * [Neach_section])
        div_points = x.xp.array(section_sizes, dtype=x.xp.intp).cumsum()

    sub_tensors = []
    for i in range(Nsections):
        st = div_points[i]
        end = div_points[i + 1]
        sub_tensors.append(x[st:end])

    return sub_tensors


def hsplit(x: Tensor, indices_or_sections: int | tuple) -> list[Tensor]:
    if not isinstance(x, Tensor):
        x = Tensor(x)

    try:
        len(indices_or_sections)
    except TypeError:
        sections = indices_or_sections
        N = x.shape[1]
        assert N % sections == 0, 'array split does not result in an equal division'

    Ntotal = x.shape[1]
    try:
        # handle array case.
        Nsections = len(indices_or_sections) + 1
        div_points = [0] + list(indices_or_sections) + [Ntotal]
    except TypeError:
        # indices_or_sections is a scalar, not an array.
        Nsections = int(indices_or_sections)
        if Nsections <= 0:
            raise ValueError(
                'number sections must be larger than 0.') from None
        Neach_section, extras = divmod(Ntotal, Nsections)
        section_sizes = ([0] + extras * [Neach_section + 1] +
                         (Nsections - extras) * [Neach_section])
        div_points = x.xp.array(section_sizes, dtype=x.xp.intp).cumsum()

    sub_tensors = []
    for i in range(Nsections):
        st = div_points[i]
        end = div_points[i + 1]
        sub_tensors.append(x[:, st:end])

    return sub_tensors


def dsplit(x: Tensor, indices_or_sections: int | tuple) -> list[Tensor]:
    if not isinstance(x, Tensor):
        x = Tensor(x)

    try:
        len(indices_or_sections)
    except TypeError:
        sections = indices_or_sections
        N = x.shape[2]
        assert N % sections == 0, 'array split does not result in an equal division'

    Ntotal = x.shape[2]
    try:
        # handle array case.
        Nsections = len(indices_or_sections) + 1
        div_points = [0] + list(indices_or_sections) + [Ntotal]
    except TypeError:
        # indices_or_sections is a scalar, not an array.
        Nsections = int(indices_or_sections)
        if Nsections <= 0:
            raise ValueError(
                'number sections must be larger than 0.') from None
        Neach_section, extras = divmod(Ntotal, Nsections)
        section_sizes = ([0] + extras * [Neach_section + 1] +
                         (Nsections - extras) * [Neach_section])
        div_points = x.xp.array(section_sizes, dtype=x.xp.intp).cumsum()

    sub_tensors = []
    for i in range(Nsections):
        st = div_points[i]
        end = div_points[i + 1]
        sub_tensors.append(x[:, :, st:end])

    return sub_tensors


def split(
    x: Tensor,
    indices_or_sections: int | tuple,
    axis: int = 0,
) -> list[Tensor]:
    if not isinstance(x, Tensor):
        x = Tensor(x)

    if axis == 0 or axis == -x.ndim:
        return vsplit(x, indices_or_sections)
    elif axis == 1 or axis == -x.ndim + 1:
        return hsplit(x, indices_or_sections)
    elif axis == 2 or axis == -x.ndim + 2:
        return dsplit(x, indices_or_sections)

    try:
        len(indices_or_sections)
    except TypeError:
        sections = indices_or_sections
        N = x.shape[axis]
        assert N % sections == 0, 'array split does not result in an equal division'

    Ntotal = x.shape[axis]
    try:
        # handle array case.
        Nsections = len(indices_or_sections) + 1
        div_points = [0] + list(indices_or_sections) + [Ntotal]
    except TypeError:
        # indices_or_sections is a scalar, not an array.
        Nsections = int(indices_or_sections)
        if Nsections <= 0:
            raise ValueError(
                'number sections must be larger than 0.') from None
        Neach_section, extras = divmod(Ntotal, Nsections)
        section_sizes = ([0] + extras * [Neach_section + 1] +
                         (Nsections - extras) * [Neach_section])
        div_points = x.xp.array(section_sizes, dtype=x.xp.intp).cumsum()

    sub_tensors = []
    stensor = swapaxes(x, 0, axis)
    for i in range(Nsections):
        st = div_points[i]
        end = div_points[i + 1]
        sub_tensors.append(swapaxes(stensor[st:end], axis, 0))
    return sub_tensors


def normalize_axis_tuple(axis, ndim, argname=None, allow_duplicate=False):
    """
    Strict public replacement for numpy.normalize_axis_tuple.

    Parameters
    ----------
    axis : None, int or iterable of int
    ndim : int
        Number of dimensions.
    argname : str, optional
        Argument name for error messages.
    allow_duplicate : bool, default False
        Whether to allow repeated axes.

    Returns
    -------
    tuple of int
    """
    if axis is None:
        return tuple(range(ndim))

    # --- convert to tuple ---
    if isinstance(axis, int):
        axis = (axis, )
    else:
        try:
            axis = tuple(axis)
        except TypeError:
            name = f"'{argname}'" if argname else "axis"
            raise TypeError(f"{name} must be int or iterable of ints")

    normalized = []

    for ax in axis:
        if not isinstance(ax, int):
            name = f"'{argname}'" if argname else "axis"
            raise TypeError(f"{name} entries must be integers")

        # --- handle negative axis ---
        if ax < 0:
            ax += ndim

        # --- bounds check ---
        if ax < 0 or ax >= ndim:
            raise ValueError(f"axis {ax - ndim if ax >= ndim else ax} "
                             f"is out of bounds for array of dimension {ndim}")

        normalized.append(ax)

    # --- duplicate check ---
    if not allow_duplicate:
        if len(set(normalized)) != len(normalized):
            raise ValueError("repeated axis in axis")

    return tuple(normalized)


def unsqueeze(x: Tensor, axis):
    '''等价于numpy的expand_dims, 因此我们借用了expand_dims的源码'''
    if type(axis) not in (tuple, list):
        axis = (axis, )

    out_ndim = len(axis) + x.ndim
    axis = normalize_axis_tuple(axis, out_ndim)

    shape_it = iter(x.shape)
    shape = [1 if ax in axis else next(shape_it) for ax in range(out_ndim)]
    return x.reshape(*shape)


def squeeze(x: Tensor, axis=None):
    shape = x.shape
    if axis is None:
        new_shape = tuple(dim for dim in shape if dim != 1)
    else:
        if isinstance(axis, int):
            axis = (axis, )
        axis = tuple(axis)

        for ax in axis:
            if ax >= len(shape) or ax < -len(shape):
                raise ValueError("Axis out of range")
            if shape[ax] != 1:
                raise ValueError(
                    f"Cannot squeeze axis {ax} with size {shape[ax]}")

        # 构造新形状，排除指定轴
        new_shape = tuple(dim for i, dim in enumerate(shape) if i not in axis)

    # 返回重塑后的数组
    return x.reshape(*new_shape)



from .module import Module
from .. import functional as F
from ...core import Tensor


class Sigmoid(Module):
    '''激活函数层 : Sigmoid'''

    def forward(self, x) -> Tensor:
        return F.sigmoid(x)

    def __repr__(self) -> str:
        return "{}()".format(self.__class__.__name__)


class Tanh(Module):
    '''激活函数层 : Tanh'''

    def forward(self, x) -> Tensor:
        return F.tanh(x)

    def __repr__(self) -> str:
        return "{}()".format(self.__class__.__name__)


class ReLU(Module):
    '''激活函数层 : ReLU'''

    def forward(self, x) -> Tensor:
        return F.relu(x)

    def __repr__(self) -> str:
        return "{}()".format(self.__class__.__name__)


class LeakyReLU(Module):
    '''
    激活函数层 : LeakyReLU
    
    Parameter
    ---------
    alpha : float
        负输入对应的斜率.
    '''

    def __init__(self, alpha: float = 0.1) -> None:
        super().__init__()
        self.alpha = float(alpha)

    def forward(self, x) -> Tensor:
        return F.leaky_relu(x, self.alpha)

    def __repr__(self) -> str:
        return "{}(alpha={})".format(self.__class__.__name__, self.alpha)


class Softmax(Module):
    '''
    激活函数层 : softmax

    Parameter
    ---------
    axis : Optional[Tuple[int]], default=None
        沿着axis计算softmax.
    '''

    def __init__(self, axis=None) -> None:
        super().__init__()
        self.axis = axis

    def forward(self, x) -> Tensor:
        return F.softmax(x, self.axis)

    def __repr__(self) -> str:
        return "{}(axis={})".format(self.__class__.__name__, self.axis)



import random
import sys
from itertools import product
from pathlib import Path

import numpy as np
import pytest

sys.path.append(str(Path(__file__).resolve().parents[1]))

import pydynet as pdn

np.random.seed(0)
random.seed(0)

type_list = [np.float16, np.float32, np.float64]


def matmul_shape_pair(max_dim=4, max_size=5):
    ndim = random.randint(0, max_dim)

    shape1 = []
    shape2 = []
    for _ in range(ndim):
        if random.random() < 0.5:
            # 50% 概率设置为 1, 确保广播可能
            s1, s2 = random.choice([(1, random.randint(1, max_size)),
                                    (random.randint(1, max_size), 1)])
        else:
            # 否则两边相同
            val = random.randint(1, max_size)
            s1, s2 = val, val
        shape1.append(s1)
        shape2.append(s2)
    shape1, shape2 = tuple(shape1), tuple(shape2)

    m = random.randint(1, max_size)
    n = random.randint(1, max_size)
    p = random.randint(1, max_size)

    shape1 = shape1 + (m, n)
    shape2 = shape2 + (n, p)

    shape1 = shape1[random.randint(0, len(shape1) - 2):]

    return shape1, shape2


def broadcastable_shape_pair(max_dim=4, max_size=5):
    ndim = random.randint(0, max_dim)  # 随机维数
    shape1 = []
    shape2 = []
    for _ in range(ndim):
        if random.random() < 0.5:
            # 50% 概率设置为 1, 确保广播可能
            s1, s2 = random.choice([(1, random.randint(1, max_size)),
                                    (random.randint(1, max_size), 1)])
        else:
            # 否则两边相同
            val = random.randint(1, max_size)
            s1, s2 = val, val
        shape1.append(s1)
        shape2.append(s2)
    shape1, shape2 = tuple(shape1), tuple(shape2)

    # 随机缺失维度
    shape1 = shape1[random.randint(0, len(shape1)):]
    return shape1, shape2


def array_pair_generator(pair_gen_func,
                         max_dim=4,
                         max_size=5,
                         n_iter=4,
                         seed=None):
    rng = np.random.default_rng(seed)
    count = 0
    while n_iter is None or count < n_iter:
        shape1, shape2 = pair_gen_func(max_dim, max_size)
        a = rng.standard_normal(size=shape1).astype(rng.choice(type_list))
        b = rng.standard_normal(size=shape2).astype(rng.choice(type_list))
        yield a, b
        count += 1


test_list = array_pair_generator(broadcastable_shape_pair, 4, 5, 8, seed=42)
func_list = [(pdn.add, np.add), (pdn.sub, np.subtract), (pdn.mul, np.multiply),
             (pdn.div, np.divide), (pdn.pow, np.power),
             (pdn.maximum, np.maximum), (pdn.minimum, np.minimum)]
test_list = [(*array, *funcs)
             for (array, funcs) in product(test_list, func_list)]


@pytest.mark.parametrize("operand1, operand2, pdn_func, np_func", test_list)
@pytest.mark.filterwarnings("ignore:invalid value")
@pytest.mark.filterwarnings("ignore:divide by zero")
def test_binary_operator(operand1: np.ndarray, operand2: np.ndarray,
                         pdn_func: callable, np_func: callable):
    pdn_operand1, pdn_operand2 = pdn.Tensor(operand1), pdn.Tensor(operand2)
    pdn_output: pdn.Tensor = pdn_func(pdn_operand1, pdn_operand2)
    np_output: np.ndarray = np_func(operand1, operand2)
    assert pdn_output.shape == np_output.shape
    assert pdn_output.dtype == np_output.dtype
    assert np.allclose(pdn_output.data, np_output, equal_nan=True)


test_list = array_pair_generator(matmul_shape_pair, 4, 5, 8, seed=42)


@pytest.mark.parametrize("operand1, operand2", test_list)
def test_matmul(operand1: np.ndarray, operand2: np.ndarray):
    pdn_operand1, pdn_operand2 = pdn.Tensor(operand1), pdn.Tensor(operand2)
    pdn_output: pdn.Tensor = pdn.matmul(pdn_operand1, pdn_operand2)
    np_output: np.ndarray = np.matmul(operand1, operand2)
    assert pdn_output.shape == np_output.shape
    assert pdn_output.dtype == np_output.dtype
    assert np.allclose(pdn_output.data, np_output, equal_nan=True)




from .optimizer import (
    Optimizer,
    SGD,
    Adagrad,
    Adadelta,
    Adam,
)
from .lr_scheduler import (
    _LRScheduler,
    ExponentialLR,
    StepLR,
    MultiStepLR,
    CosineAnnealingLR,
)



from .tensor import (Tensor, add, sub, mul, div, pow, matmul, abs, sum, mean,
                     min, max, min, argmax, argmin, maximum, minimum, exp, log,
                     sign, reshape, transpose, swapaxes, concat, sigmoid, tanh)
from .function import sqrt, square, vsplit, hsplit, dsplit, split, unsqueeze, squeeze



import numpy as np

from ..core import tensor, function
from .. import unsqueeze, no_grad


def linear(x: tensor.Tensor, weight: tensor.Tensor, bias: tensor.Tensor):
    affine = x @ weight
    if bias is not None:
        affine = affine + bias
    return affine


def embedding(x: tensor.Tensor, weight: tensor.Tensor, padding_idx: int):
    query = weight[x]
    if padding_idx is not None:
        with tensor.no_grad():
            mask = unsqueeze(x.ne(padding_idx), -1)
        query = query * mask
    return query


def sigmoid(x: tensor.Tensor):
    return tensor.sigmoid(x)


def tanh(x: tensor.Tensor):
    return tensor.tanh(x)


def relu(x: tensor.Tensor):
    return tensor.maximum(0., x)


def leaky_relu(x: tensor.Tensor, alpha: float):
    return tensor.maximum(x, alpha * x)


def silu(x: tensor.Tensor):
    return x / (1 + tensor.exp(-x))


def softmax(x: tensor.Tensor, axis=None):
    '''Softmax函数'''
    with no_grad():
        max_ = x.max(axis, keepdims=True)
    x_sub_max = x - max_
    exp_ = tensor.exp(x_sub_max)
    return exp_ / tensor.sum(exp_, axis=axis, keepdims=True)


def log_softmax(x: tensor.Tensor, axis=None, keepdims=False):
    '''log-softmax函数'''
    with no_grad():
        max_ = x.max(axis, keepdims=True)
    x_sub_max = x - max_
    return x_sub_max - tensor.log(
        tensor.sum(tensor.exp(x_sub_max), axis=axis, keepdims=keepdims))


class __im2col1d(tensor._UnaryOperator):

    def __init__(
        self,
        x: tensor.Tensor,
        kernel_size: int,
        stride: int,
    ) -> None:
        self.N, self.in_channels, self.n_features = x.shape
        self.kernel_size = kernel_size
        self.stride = stride
        self.n_output = (self.n_features - self.kernel_size) // stride + 1
        super().__init__(x)

    def forward_(self, x: tensor.Tensor) -> np.ndarray:
        s0, s1, s2 = x.strides
        shape = (x.shape[0], self.in_channels, self.kernel_size, self.n_output)
        self.__strides = (s0, s1, s2, s2 * self.stride)

        col = self.xp.lib.stride_tricks.as_strided(
            x.data,
            shape=shape,
            strides=self.__strides,
        ).copy()
        return col

    def grad_fn(self, x: tensor.Tensor, grad: np.ndarray) -> np.ndarray:
        grad_x = self.xp.zeros(x.shape, dtype=self.dtype)
        view = self.xp.lib.stride_tricks.as_strided(
            grad_x,
            shape=self.shape,
            strides=self.__strides,
        )
        self.xp.add.at(view, (..., ), grad)
        return grad_x


class __pad1d(tensor._UnaryOperator):

    def __init__(self, x: tensor.Tensor, pad_width=0) -> None:
        self.pad_width = pad_width
        super().__init__(x)

    def forward_(self, x: tensor.Tensor) -> np.ndarray:
        return self.xp.pad(x.data, [(0, 0), (0, 0),
                                    (self.pad_width, self.pad_width)],
                           'constant')

    def grad_fn(self, x: tensor.Tensor, grad: np.ndarray) -> np.ndarray:
        if self.pad_width == 0:
            return grad[...]
        return grad[..., self.pad_width:-self.pad_width]


def conv1d(
    x: tensor.Tensor,
    kernel: tensor.Tensor,
    padding: int = 0,
    stride: int = 1,
):
    '''一维卷积函数

    基于im2col实现的一维卷积.
    
    Parameters
    ----------
    x : Tensor
        输入数据, 形状为(N, in_channels, n_features);
    kernel : Tensor
        卷积核, 形状为(out_channels, in_channels, kernel_size);
    padding : int, default=0
        对输入特征两边补0数量;
    stride : int, default=1
        卷积步长.
    '''
    kernel_size = kernel.shape[-1]
    pad_x = __pad1d(x, padding)
    col = __im2col1d(pad_x, kernel_size, stride)
    return (col @ kernel.transpose(1, 2, 0)).sum(1).swapaxes(1, 2)


def max_pool1d(
    x: tensor.Tensor,
    kernel_size: int,
    stride: int,
    padding: int = 0,
):
    '''一维池化函数

    基于im2col实现的一维池化.`
    
    Parameters
    ----------
    x : Tensor
        输入数据, 形状为(N, in_channels, n_features);
    kernel_size : int
        池化核大小;
    stride : int
        卷积步长;
    padding : int, default=0
        对输入特征两边补0数量.
    '''
    pad_x = __pad1d(x, padding)
    col = __im2col1d(pad_x, kernel_size, stride)
    return col.max(-1)


def avg_pool1d(
    x: tensor.Tensor,
    kernel_size: int,
    stride: int,
    padding: int = 0,
):
    '''一维平均池化函数

    基于im2col实现的一维池化.`
    
    Parameters
    ----------
    x : Tensor
        输入数据, 形状为(N, in_channels, n_features);
    kernel_size : int
        池化核大小;
    stride : int
        卷积步长;
    padding : int, default=0
        对输入特征两边补0数量.
    '''
    pad_x = __pad1d(x, padding)
    col = __im2col1d(pad_x, kernel_size, stride)
    return col.mean(-1)


class __im2col2d(tensor._UnaryOperator):

    def __init__(
        self,
        x: tensor.Tensor,
        kernel_size: int,
        stride: int,
    ) -> None:
        _, self.in_channels, self.n_h, self.n_w = x.shape
        self.kernel_size = kernel_size
        self.stride = stride
        self.out_h, self.out_w = (
            self.n_h - self.kernel_size) // self.stride + 1, (
                self.n_w - self.kernel_size) // self.stride + 1

        super().__init__(x)

    def forward_(self, x: tensor.Tensor) -> np.ndarray:
        s0, s1, s2, s3 = x.strides
        shape = (x.shape[0], self.in_channels, self.kernel_size,
                 self.kernel_size, self.out_h, self.out_w)
        self.__strides = (s0, s1, s2, s3, s2 * self.stride, s3 * self.stride)

        col = self.xp.lib.stride_tricks.as_strided(
            x.data,
            shape=shape,
            strides=self.__strides,
        ).copy()
        return col

    def grad_fn(self, x: tensor.Tensor, grad: np.ndarray) -> np.ndarray:
        grad_x = self.xp.zeros(x.shape, dtype=self.dtype)
        view = self.xp.lib.stride_tricks.as_strided(
            grad_x,
            shape=self.shape,
            strides=self.__strides,
        )
        self.xp.add.at(view, (..., ), grad)
        return grad_x


class __pad2d(tensor._UnaryOperator):

    def __init__(self, x: tensor.Tensor, pad_width=0) -> None:
        self.pad_width = pad_width
        super().__init__(x)

    def forward_(self, x: tensor.Tensor) -> np.ndarray:
        return self.xp.pad(x.data, [(0, 0), (0, 0),
                                    (self.pad_width, self.pad_width),
                                    (self.pad_width, self.pad_width)],
                           'constant')

    def grad_fn(self, x: tensor.Tensor, grad: np.ndarray) -> np.ndarray:
        if self.pad_width == 0:
            return grad[...]
        return grad[..., self.pad_width:-self.pad_width,
                    self.pad_width:-self.pad_width]


def conv2d(x: tensor.Tensor,
           kernel: tensor.Tensor,
           padding: int = 0,
           stride: int = 1):
    '''二维卷积函数

    基于im2col实现的二维卷积. 为了实现上的方便, 我们不考虑长宽不同的卷积核, 步长和补零。
    
    Parameters
    ----------
    x : Tensor
        输入数据, 形状为(N, in_channels, n_height, n_width);
    kernel : Tensor
        卷积核, 形状为(out_channels, in_channels, kernel_height, kernel_width);
    padding : int, default=0
        对输入图片周围补0数量;
    stride : int, default=1
        卷积步长.
    '''
    N, _, _, _ = x.shape
    out_channels, _, kernel_size, _ = kernel.shape
    pad_x = __pad2d(x, padding)
    col = __im2col2d(pad_x, kernel_size, stride)
    out_h, out_w = col.shape[-2:]
    col = col.transpose(0, 4, 5, 1, 2, 3).reshape(N * out_h * out_w, -1)
    col_filter = kernel.reshape(out_channels, -1).T
    out = col @ col_filter
    return out.reshape(N, out_h, out_w, -1).transpose(0, 3, 1, 2)


def max_pool2d(x: tensor.Tensor, kernel_size: int, stride: int, padding=0):
    '''二维卷积函数池化

    基于im2col实现的二维卷积. 为了实现上的方便, 我们不考虑长宽不同的kernel_size, 步长和补零。
    
    Parameters
    ----------
    x : Tensor
        输入数据, 形状为(N, in_channels, n_height, n_width);
    kernel_size : int
        池化核尺寸;
    stride : int, default=1
        卷积步长;
    padding : int, default=0
        对输入图片周围补0数量;
    '''
    N, in_channels, _, _ = x.shape
    pad_x = __pad2d(x, padding)
    col = __im2col2d(pad_x, kernel_size, stride)
    out_h, out_w = col.shape[-2:]
    col = col.transpose(0, 4, 5, 1, 2, 3).reshape(
        -1,
        kernel_size * kernel_size,
    )
    out = col.max(1)
    out = out.reshape(N, out_h, out_w, in_channels).transpose(0, 3, 1, 2)
    return out


def avg_pool2d(x: tensor.Tensor, kernel_size: int, stride: int, padding=0):
    '''二维平均池化

    基于im2col实现的二维池化. 为了实现上的方便, 我们不考虑长宽不同的kernel_size, 步长和补零。
    
    Parameters
    ----------
    x : Tensor
        输入数据, 形状为(N, in_channels, n_height, n_width);
    kernel_size : int
        池化核尺寸;
    stride : int, default=1
        卷积步长;
    padding : int, default=0
        对输入图片周围补0数量;
    '''
    N, in_channels, _, _ = x.shape
    pad_x = __pad2d(x, padding)
    col = __im2col2d(pad_x, kernel_size, stride)
    out_h, out_w = col.shape[-2:]
    col = col.transpose(0, 4, 5, 1, 2, 3).reshape(
        -1,
        kernel_size * kernel_size,
    )
    out = col.mean(1)
    out = out.reshape(N, out_h, out_w, in_channels).transpose(0, 3, 1, 2)
    return out


def mse_loss(y_pred, y_true, reduction='mean'):
    '''均方误差'''
    square_sum = function.square(y_pred - y_true)
    if reduction == 'mean':
        return tensor.mean(square_sum)
    elif reduction == 'sum':
        return tensor.sum(square_sum)
    else:
        raise ValueError("reduction must be mean or sum.")


def nll_loss(y_pred, y_true, reduction='mean'):
    '''负对数似然'''
    nll = -y_pred * y_true
    if reduction == 'mean':
        return tensor.mean(nll)
    elif reduction == 'sum':
        return tensor.sum(nll)
    else:
        raise ValueError("reduction must be mean or sum.")


def cross_entropy_loss(y_pred, y_true, reduction='mean'):
    '''交叉熵损失'''
    update_y_pred = y_pred - y_pred.max().item()
    log_sum_exp = tensor.log(
        tensor.sum(tensor.exp(update_y_pred), 1, keepdims=True))

    neg_log_sm = log_sum_exp - update_y_pred
    if y_true.ndim == 1:
        nll = neg_log_sm[range(len(neg_log_sm)), y_true]
    else:
        nll = neg_log_sm * y_true

    if reduction == 'mean':
        return tensor.mean(nll)
    elif reduction == 'sum':
        return tensor.sum(nll)
    else:
        raise ValueError("reduction must be mean or sum.")



import functools

grad_enable = True


def is_grad_enable():
    return grad_enable


def set_grad_enabled(mode: bool):
    global grad_enable
    grad_enable = mode


class no_grad:

    def __enter__(self) -> None:
        self.prev = is_grad_enable()
        set_grad_enabled(False)

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        set_grad_enabled(self.prev)

    def __call__(self, func):

        @functools.wraps(func)
        def decorate_context(*args, **kwargs):
            with __class__():
                return func(*args, **kwargs)

        return decorate_context


class enable_grad:

    def __enter__(self) -> None:
        self.prev = is_grad_enable()
        set_grad_enabled(True)

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        set_grad_enabled(self.prev)

    def __call__(self, func):

        @functools.wraps(func)
        def decorate_context(*args, **kwargs):
            with __class__():
                return func(*args, **kwargs)

        return decorate_context



from .core import (Tensor, add, sub, mul, div, pow, matmul, abs, sum, mean,
                   min, max, argmax, argmin, maximum, minimum, exp, log,
                   sign, reshape, transpose, swapaxes, concat, sigmoid, tanh,
                   sqrt, square, vsplit, hsplit, dsplit, split, unsqueeze,
                   squeeze)
from .special import zeros, ones, rand, randn, empty, uniform
from .cuda import Device
from .autograd import enable_grad, no_grad

__all__ = [
    "Tensor", "add", "sub", "mul", "div", "pow", "matmul", "abs", "sum",
    "mean", "min", "max", "argmax", "argmin", "maximum", "minimum", "exp",
    "log", "sign", "reshape", "transpose", "swapaxes", "concat", 'sigmoid',
    'tanh', "sqrt", "square", "vsplit", "hsplit", "dsplit", "split",
    "unsqueeze", "squeeze", "zeros", "ones", "rand", "randn", "empty",
    "uniform", "Device", "enable_grad", "no_grad"
]



from .module import Module
from ..parameter import Parameter
from .. import init, functional as F
from ...core import Tensor
from ...special import empty
from ...cuda import Device
from ...autograd import no_grad

import math


class Linear(Module):

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        device=None,
        dtype=None,
    ) -> None:
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        kwargs = {"device": Device(device), "dtype": dtype}
        self.weight = Parameter(
            empty((self.in_features, self.out_features), **kwargs))
        self.bias = Parameter(empty(self.out_features, **
                                    kwargs)) if bias else None
        self.reset_paramters()

    def reset_paramters(self):
        init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = init._calculate_fan(self.weight)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            init.uniform_(self.bias, -bound, bound)

    def forward(self, x: Tensor):
        return F.linear(x, self.weight, self.bias)

    def __repr__(self) -> str:
        return "Linear(in_features={}, out_features={}, bias={})".format(
            self.in_features, self.out_features, self.bias is not None)


class Embedding(Module):

    def __init__(
        self,
        num_embeddings: int,
        embedding_dim: int,
        padding_idx=None,
        device=None,
        dtype=None,
    ) -> None:
        super().__init__()
        self.num_embedding = num_embeddings
        self.embedding_dim = embedding_dim
        self.padding_idx = padding_idx

        kwargs = {"device": Device(device), "dtype": dtype}
        self.weight = Parameter(
            empty((self.num_embedding, self.embedding_dim), **kwargs))

    def forward(self, x: Tensor):
        return F.embedding(x, self.weight, self.padding_idx)

    def reset_parameters(self) -> None:
        init.normal_(self.weight)
        self._fill_padding_idx_with_zero()

    def _fill_padding_idx_with_zero(self) -> None:
        if self.padding_idx is not None:
            with no_grad():
                self.weight[self.padding_idx].data = self.weight.xp.zeros(
                    self.weight[self.padding_idx].shape,
                    dtype=self.weight.dtype,
                )



import numpy as np
from .core import Tensor


# 一些包装的特殊矩阵
def zeros(shape, dtype=None, device=None, requires_grad=False):
    '''全0张量
    
    Parameters
    ----------
    shape : 
        张量形状
    require_grad : bool, default=False
        是否需要求导
    '''
    return Tensor(np.zeros(shape),
                  dtype=dtype,
                  device=device,
                  requires_grad=requires_grad)


def ones(shape, dtype=None, device=None, requires_grad=False):
    '''全1张量
    
    Parameters
    ----------
    shape : 
        张量形状
    require_grad : bool, default=False
        是否需要求导
    '''
    return Tensor(np.ones(shape),
                  dtype=dtype,
                  device=device,
                  requires_grad=requires_grad)


def randn(*shape, dtype=None, device=None, requires_grad=False):
    '''0-1正态分布张量
    
    Parameters
    ----------
    *shape : 
        张量形状
    require_grad : bool, default=False
        是否需要求导
    '''
    return Tensor(np.random.randn(*shape),
                  dtype=dtype,
                  device=device,
                  requires_grad=requires_grad)


def rand(*shape, dtype=None, device=None, requires_grad=False):
    '''[0, 1)均匀分布张量
    
    Parameters
    ----------
    *shape : 
        张量形状
    require_grad : bool, default=False
        是否需要求导
    '''
    return Tensor(np.random.rand(*shape),
                  dtype=dtype,
                  device=device,
                  requires_grad=requires_grad)


def uniform(low: float,
            high: float,
            shape=None,
            dtype=None,
            device=None,
            requires_grad=False):
    '''均匀分布张量
    
    Parameters
    ----------
    low : float
        均匀分布下界;
    high : float
        均匀分布下界;
    *shape : 
        张量形状
    require_grad : bool, default=False
        是否需要求导
    '''
    return Tensor(np.random.uniform(low, high, size=shape),
                  dtype=dtype,
                  device=device,
                  requires_grad=requires_grad)


def empty(shape, dtype=None, device=None, requires_grad=False):
    return Tensor(np.empty(shape, dtype=dtype),
                  dtype=dtype,
                  device=device,
                  requires_grad=requires_grad)



from .module import Module
from .. import functional as F


class MaxPool1d(Module):

    def __init__(self, kernel_size: int, stride: int, padding: int) -> None:
        super().__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding

    def forward(self, x):
        return F.max_pool1d(x, self.kernel_size, self.stride, self.padding)

    def __repr__(self) -> str:
        return "{}(kernel_size={}, stride={}, padding={})".format(
            self.__class__.__name__,
            self.kernel_size,
            self.stride,
            self.padding,
        )


class AvgPool1d(Module):

    def __init__(self, kernel_size: int, stride: int, padding: int) -> None:
        super().__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding

    def forward(self, x):
        return F.avg_pool1d(x, self.kernel_size, self.stride, self.padding)

    def __repr__(self) -> str:
        return "{}(kernel_size={}, stride={}, padding={})".format(
            self.__class__.__name__,
            self.kernel_size,
            self.stride,
            self.padding,
        )


class MaxPool2d(Module):

    def __init__(self, kernel_size: int, stride: int, padding: int) -> None:
        super().__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding

    def forward(self, x):
        return F.max_pool2d(x, self.kernel_size, self.stride, self.padding)

    def __repr__(self) -> str:
        return "{}(kernel_size={}, stride={}, padding={})".format(
            self.__class__.__name__,
            self.kernel_size,
            self.stride,
            self.padding,
        )


class AvgPool2d(Module):

    def __init__(self, kernel_size: int, stride: int, padding: int) -> None:
        super().__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding

    def forward(self, x):
        return F.avg_pool2d(x, self.kernel_size, self.stride, self.padding)

    def __repr__(self) -> str:
        return "{}(kernel_size={}, stride={}, padding={})".format(
            self.__class__.__name__,
            self.kernel_size,
            self.stride,
            self.padding,
        )



from .module import Module
from ...core import Tensor
from ...special import rand


class Dropout(Module):

    def __init__(self, p: float = 0.5) -> None:
        super().__init__()
        assert p >= 0 and p < 1
        self.p = p

    def forward(self, x) -> Tensor:
        if self._train:
            mask = rand(*x.shape, device=x.device) < 1 - self.p
            return x * mask.astype(x.dtype) / (1 - self.p)
        return x

    def __repr__(self) -> str:
        return "{}(p={})".format(self.__class__.__name__, self.p)



from .module import Module
from ..parameter import Parameter
from .. import init
from .. import functional as F
from ...special import empty
from ...cuda import Device

import math


class Conv1d(Module):

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        padding: int = 0,
        bias: bool = True,
        device=None,
        dtype=None,
    ) -> None:
        super().__init__()
        kwargs = {"device": Device(device), "dtype": dtype}
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.padding = padding
        self.stride = stride
        self.weight = Parameter(
            empty((self.out_channels, self.in_channels, self.kernel_size),
                  **kwargs))
        self.bias = Parameter(empty(
            (1, self.out_channels, 1), **kwargs)) if bias else None
        self.reset_parameters()

    def reset_parameters(self):
        init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = init._calculate_fan(self.weight)
            if fan_in != 0:
                bound = 1 / math.sqrt(fan_in)
                init.uniform_(self.bias, -bound, bound)

    def forward(self, x):
        conv1d = F.conv1d(x, self.weight, self.padding, self.stride)
        if self.bias is not None:
            return conv1d + self.bias
        return conv1d

    def __repr__(self) -> str:
        return "{}(in_channels={}, out_channels={}, kernel_size={}, padding={}, stride={}, bias={})".format(
            self.__class__.__name__,
            self.in_channels,
            self.out_channels,
            self.kernel_size,
            self.padding,
            self.stride,
            self.bias is not None,
        )


class Conv2d(Module):

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        padding: int = 0,
        bias: bool = True,
        device=None,
        dtype=None,
    ) -> None:
        super().__init__()
        kwargs = {"device": Device(device), "dtype": dtype}
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.padding = padding
        self.stride = stride
        self.weight = Parameter(
            empty((self.out_channels, self.in_channels, self.kernel_size,
                   self.kernel_size), **kwargs))
        self.bias = Parameter(empty(
            (1, self.out_channels, 1, 1), **kwargs)) if bias else None
        self.reset_parameters()

    def reset_parameters(self):
        init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = init._calculate_fan(self.weight)
            if fan_in != 0:
                bound = 1 / math.sqrt(fan_in)
                init.uniform_(self.bias, -bound, bound)

    def forward(self, x):
        conv2d = F.conv2d(x, self.weight, self.padding, self.stride)
        if self.bias is not None:
            return conv2d + self.bias
        return conv2d

    def __repr__(self) -> str:
        return "{}(in_channels={}, out_channels={}, kernel_size={}, padding={}, stride={}, bias={})".format(
            self.__class__.__name__,
            self.in_channels,
            self.out_channels,
            self.kernel_size,
            self.padding,
            self.stride,
            self.bias is not None,
        )



from numpy.random import permutation


class Dataset:

    def __init__(self) -> None:
        pass

    def __getitem__(self, index):
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError


class Sampler:

    def __init__(self, dataset: Dataset) -> None:
        pass

    def __iter__(self):
        raise NotImplementedError


class SequentialSampler(Sampler):

    def __init__(self, dataset: Dataset) -> None:
        self.dataset = dataset

    def __iter__(self):
        return iter(range(len(self.dataset)))

    def __len__(self) -> int:
        return len(self.dataset)


class RandomSampler(Sampler):

    def __init__(self, dataset: Dataset) -> None:
        self.dataset = dataset

    def __iter__(self):
        yield from permutation(len(self.dataset)).tolist()

    def __len__(self):
        return len(self.dataset)


class BatchSampler(Sampler):

    def __init__(self, sampler: Sampler, batch_size: int,
                 drop_last: bool) -> None:
        self.sampler = sampler
        self.batch_size = batch_size
        self.drop_last = drop_last

    def __iter__(self):
        batch = []
        for idx in self.sampler:
            batch.append(idx)
            if len(batch) == self.batch_size:
                yield batch
                batch = []
        if len(batch) > 0 and not self.drop_last:
            yield batch

    def __len__(self):
        if self.drop_last:
            return len(self.sampler) // self.batch_size
        return (len(self.sampler) + self.batch_size - 1) // self.batch_size


class _DataLoaderIter:

    def __init__(self, loader) -> None:
        self.loader = loader
        self.sample_iter = iter(self.loader.batch_sampler)

    def __next__(self):
        index = next(self.sample_iter)
        return self.loader.dataset[index]


class DataLoader:

    def __init__(self,
                 dataset: Dataset,
                 batch_size: int = 1,
                 shuffle: bool = False,
                 drop_last: bool = False) -> None:
        self.dataset = dataset
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.drop_last = drop_last

        if shuffle:
            self.sampler = RandomSampler(dataset)
        else:
            self.sampler = SequentialSampler(dataset)

        self.batch_sampler = BatchSampler(self.sampler, batch_size, drop_last)
        self.batch_size = batch_size
        self.drop_last = drop_last

    def __iter__(self):
        return _DataLoaderIter(self)


def data_loader(X, y, batch_size: int, shuffle: bool = False) -> list:

    class TrainSet(Dataset):

        def __init__(self, X, y) -> None:
            self.data = X
            self.target = y

        def __getitem__(self, index):
            return self.data[index], self.target[index]

        def __len__(self):
            return len(self.data)

    return DataLoader(TrainSet(X, y), batch_size, shuffle)



import sys
from pathlib import Path

import numpy as np
import pytest

sys.path.append(str(Path(__file__).resolve().parents[1]))

import pydynet as pdn


np.random.seed(1)


def test_unary_function_forward_matches_numpy():
    x_np = np.random.uniform(0.5, 2.0, size=(3, 4)).astype(np.float64)
    x = pdn.Tensor(x_np)

    pairs = [
        (pdn.abs, np.abs),
        (pdn.exp, np.exp),
        (pdn.log, np.log),
        (pdn.sign, np.sign),
        (pdn.sigmoid, lambda z: 1.0 / (1.0 + np.exp(-z))),
        (pdn.tanh, np.tanh),
        (pdn.sqrt, np.sqrt),
        (pdn.square, np.square),
    ]

    for pdn_func, np_func in pairs:
        pdn_out = pdn_func(x)
        np_out = np_func(x_np)
        assert pdn_out.shape == np_out.shape
        assert np.allclose(pdn_out.data, np_out, atol=1e-6, rtol=1e-6)


def test_reduce_function_forward_matches_numpy():
    x_np = np.random.randn(2, 3, 4).astype(np.float64)
    x = pdn.Tensor(x_np)

    test_cases = [
        (lambda t: pdn.sum(t), lambda a: np.sum(a)),
        (lambda t: pdn.mean(t), lambda a: np.mean(a)),
        (lambda t: pdn.sum(t, axis=1), lambda a: np.sum(a, axis=1)),
        (lambda t: pdn.mean(t, axis=(0, 2), keepdims=True),
         lambda a: np.mean(a, axis=(0, 2), keepdims=True)),
        (lambda t: pdn.max(t, axis=2), lambda a: np.max(a, axis=2)),
        (lambda t: pdn.min(t, axis=0), lambda a: np.min(a, axis=0)),
        (lambda t: pdn.argmax(t, axis=1), lambda a: np.argmax(a, axis=1)),
        (lambda t: pdn.argmin(t, axis=2), lambda a: np.argmin(a, axis=2)),
    ]

    for pdn_func, np_func in test_cases:
        pdn_out = pdn_func(x)
        np_out = np_func(x_np)
        assert pdn_out.shape == np_out.shape
        assert np.allclose(pdn_out.data, np_out)


def test_shape_manipulation_matches_numpy():
    x_np = np.arange(24, dtype=np.float64).reshape(2, 3, 4)
    x = pdn.Tensor(x_np)

    reshape_out = pdn.reshape(x, (4, 6))
    assert np.array_equal(reshape_out.data, x_np.reshape(4, 6))

    transpose_out = pdn.transpose(x, (2, 0, 1))
    assert np.array_equal(transpose_out.data, x_np.transpose(2, 0, 1))

    swapaxes_out = pdn.swapaxes(x, 0, 2)
    assert np.array_equal(swapaxes_out.data, np.swapaxes(x_np, 0, 2))

    unsqueeze_out = pdn.unsqueeze(x, (0, 2))
    assert np.array_equal(unsqueeze_out.data, np.expand_dims(np.expand_dims(x_np, 0), 2))

    squeeze_in = pdn.Tensor(np.ones((1, 2, 1, 3), dtype=np.float64))
    squeeze_out = pdn.squeeze(squeeze_in, axis=(0, 2))
    assert np.array_equal(squeeze_out.data, np.squeeze(squeeze_in.data, axis=(0, 2)))


@pytest.mark.parametrize("axis", [0, 1, 2])
def test_split_and_concat_roundtrip(axis):
    x_np = np.random.randn(4, 6, 8).astype(np.float64)
    x = pdn.Tensor(x_np)

    pieces = pdn.split(x, 2, axis=axis)
    assert len(pieces) == 2

    merged = pdn.concat(pieces, axis=axis)
    assert np.allclose(merged.data, x_np)


def test_concat_backward_distributes_gradient():
    a_np = np.random.randn(2, 3).astype(np.float64)
    b_np = np.random.randn(2, 2).astype(np.float64)

    a = pdn.Tensor(a_np, requires_grad=True)
    b = pdn.Tensor(b_np, requires_grad=True)

    y = pdn.concat([a, b], axis=1).sum()
    y.backward()

    assert np.array_equal(a.grad, np.ones_like(a_np))
    assert np.array_equal(b.grad, np.ones_like(b_np))


def test_mean_backward_with_axis_and_keepdims():
    x_np = np.random.randn(2, 3, 4).astype(np.float64)
    x = pdn.Tensor(x_np, requires_grad=True)

    y = pdn.mean(x, axis=1, keepdims=True).sum()
    y.backward()

    expected = np.ones_like(x_np) / x_np.shape[1]
    assert np.allclose(x.grad, expected)



from .module import Module
from .. import functional as F
from ...core import Tensor


class Loss(Module):
    '''损失函数基类'''

    def __init__(self, reduction='mean') -> None:
        super().__init__()
        self.reduction = reduction
        assert self.reduction in {'mean', 'sum'}

    def forward(self, y_pred: Tensor, y_true: Tensor) -> Tensor:
        raise NotImplementedError


class MSELoss(Loss):

    def forward(self, y_pred: Tensor, y_true: Tensor) -> Tensor:
        return F.mse_loss(y_pred, y_true, reduction=self.reduction)


class NLLLoss(Loss):

    def forward(self, y_pred: Tensor, y_true: Tensor) -> Tensor:
        return F.nll_loss(y_pred, y_true, reduction=self.reduction)


class CrossEntropyLoss(Loss):

    def forward(self, y_pred: Tensor, y_true: Tensor) -> Tensor:
        return F.cross_entropy_loss(y_pred, y_true, reduction=self.reduction)



'''学习率调节器类，我们目前实现了\n
- ExponentialLR;\n
- StepLR;\n
- MultiStepLR;\n
- CosineAnnealingLR.\n
'''

from typing import List
from .optimizer import Optimizer
import weakref
from functools import wraps
from collections import Counter
from math import cos, pi


class _LRScheduler:
    def __init__(self, optimizer: Optimizer, last_epoch: int = -1) -> None:
        self.optimizer = optimizer
        self.last_epoch = last_epoch

        if self.last_epoch == -1:
            self.optimizer.initial_lr = self.optimizer.lr
        else:
            assert hasattr(
                self.optimizer, "initial_lr"
            ), "last_epoch=1 but no 'initial_lr' attribute in optimizer!"

        def with_counter(method):
            if getattr(method, '_with_counter', False):
                # `optimizer.step()` has already been replaced, return.
                return method

            # 建立一个method的弱引用。弱引用不增加对象的引用计数,只存在弱引用的对象是可被垃圾回收的;
            # 弱引用可以解决循环引用的问题。
            instance_ref = weakref.ref(method.__self__)
            # Get the unbound method for the same purpose.
            func = method.__func__  # __func__是method的底层实现,不跟具体的实例绑定
            cls = instance_ref().__class__  # method的所属类
            del method

            @wraps(func)
            def wrapper(*args, **kwargs):
                instance = instance_ref()
                instance._step_count += 1
                wrapped = func.__get__(instance, cls)
                return wrapped(*args, **kwargs)

            # Note that the returned function here is no longer a bound method,
            # so attributes like `__func__` and `__self__` no longer exist.
            wrapper._with_counter = True
            return wrapper

        # 通过装饰器来为optimizer.step添加计数功能,并初始化计数器
        self.optimizer.step = with_counter(self.optimizer.step)
        self.optimizer._step_count = 0
        self._step_count = 0

        self.step()

    def step(self):
        self._step_count += 1  # lr_scheduler的step计数

        # 支持上下文管理器协议的类
        class _enable_get_lr_call:
            def __init__(self, o):
                self.o = o

            def __enter__(self):
                self.o._get_lr_called_within_step = True
                return self

            def __exit__(self, type, value, traceback):
                self.o._get_lr_called_within_step = False

        with _enable_get_lr_call(self):
            self.last_epoch += 1  # 更新epoch
            lr = self.get_lr()  # 计算新的lr,与具体的lr_scheduler类型有关

        # _last_lr记录上一轮次更新的lr值
        self._last_lr = self.optimizer.lr
        self.optimizer.lr = lr

    def get_lr(self):
        raise NotImplementedError

    def get_last_lr(self):
        return self._last_lr


class ExponentialLR(_LRScheduler):
    def __init__(
        self,
        optimizer: Optimizer,
        gamma: float = 0.1,
        last_epoch: int = -1,
    ) -> None:
        self.gamma = gamma
        super().__init__(optimizer, last_epoch)

    def get_lr(self):
        return self.optimizer.lr * self.gamma**self.last_epoch


class StepLR(_LRScheduler):
    def __init__(
        self,
        optimizer: Optimizer,
        step_size: int,
        gamma=0.1,
        last_epoch: int = -1,
    ) -> None:
        self.step_size = step_size
        self.gamma = gamma
        super().__init__(optimizer, last_epoch)

    def get_lr(self):
        return self.optimizer.lr * self.gamma**(self.last_epoch //
                                                self.step_size)


class MultiStepLR(_LRScheduler):
    def __init__(
        self,
        optimizer: Optimizer,
        milestones: List[int],
        gamma=0.1,
        last_epoch: int = -1,
    ) -> None:
        self.milestones = Counter(milestones)
        self.gamma = gamma
        super().__init__(optimizer, last_epoch)

    def get_lr(self):
        if self.last_epoch not in self.milestones:
            return self.optimizer.lr
        return self.optimizer.lr * self.gamma**self.milestones[self.last_epoch]


class CosineAnnealingLR(_LRScheduler):
    def __init__(
        self,
        optimizer: Optimizer,
        T_max: int,
        eta_min: float = 0,
        last_epoch: int = -1,
    ) -> None:
        self.T_max = T_max
        self.eta_min = eta_min
        super().__init__(optimizer, last_epoch)

    def get_lr(self):
        base_lr = self.optimizer.initial_lr
        if self.last_epoch == 0:
            return base_lr
        elif (self.last_epoch - 1 - self.T_max) % (2 * self.T_max) == 0:
            return self.get_last_lr() + (base_lr - self.eta_min) * (
                1 - cos(pi / self.T_max)) / 2
        return (1 + cos(pi * self.last_epoch / self.T_max)) / (
            1 + cos(pi * (self.last_epoch - 1) / self.T_max)) * (
                self.get_last_lr() - self.eta_min) + self.eta_min



from __future__ import annotations
import numpy as np
import numpy.typing as npt

from ..cuda import Device
from ..autograd import is_grad_enable, no_grad


class Graph:
    '''计算图, 全局共用一个动态计算图'''
    node_list: list[Tensor] = list()
    size = 0

    @classmethod
    def _add_node(cls, node):
        '''添加图节点'''
        cls.node_list.append(node)
        cls.size += 1

    @classmethod
    def _free_node(cls, node):
        node.last.clear()

        index = cls.node_list.index(node)
        cls.node_list.pop(index)

        cls.size -= 1


class Tensor:
    '''
    将数据(NumPy数组)包装成可微分张量

    Parameters
    ----------
    data : ndarray
        张量数据, 只要是np.array能够转换的数据;
    dtype : default=None
        数据类型, 和numpy数组的dtype等价
    copy : bool, default=False
        是否拷贝数据
    requires_grad : bool, default=False
        是否需要求梯度;

    Attributes
    ----------
    data : numpy.ndarray
        核心数据, 为NumPy数组;
    requires_grad : bool
        是否需要求梯度;
    grad : numpy.ndarray
        梯度数据, 为和data相同形状的数组(初始化为全0);
    last : list[Tensor]
        上游节点列表.

    Example
    -------
    >>> import numpy as np
    >>> from pydynet.tensor import Tensor
    >>> x = Tensor(1., requires_grad=True)
    >>> y = Tensor([1, 2, 3], dtype=float)
    >>> z = Tensor(np.random.rand(3, 4))
    '''

    def __init__(
        self,
        data: npt.ArrayLike,
        dtype: npt.DTypeLike = None,
        copy: bool = True,
        device: Device | int | str | None = None,
        requires_grad: bool = False,
    ) -> None:
        if isinstance(data, Tensor):
            raise ValueError(
                'Tensor assignment with another tensor is forbidden.')

        self.copy = copy
        self.device = Device(device)
        with self.device:
            self.data = self.xp.array(data, dtype=dtype, copy=self.copy)
            # self.data = self.xp.asarray(data, dtype=dtype)

        self.requires_grad = is_grad_enable() and requires_grad
        if self.requires_grad:
            if not self.xp.issubdtype(self.dtype, np.floating):
                raise TypeError(
                    "Only Tensors of floating point dtype can require gradients!"
                )
            with self.device:
                self.grad = self.xp.zeros(self.shape, dtype=dtype)
            self.last: list[Tensor] = list()
            Graph._add_node(self)
        else:
            self.grad = None

    @property
    def is_leaf(self) -> bool:
        '''判断是否为叶节点:需要求导且无上游节点的节点为叶节点.'''
        return not self.requires_grad or len(self.last) == 0

    @property
    def shape(self) -> tuple[int]:
        '''张量的形状, 用法同NumPy.
        
        Example
        -------
        >>> from pydynet import Tensor
        >>> Tensor([[2, 2]]).shape
        (1, 2)
        '''
        return self.data.shape

    @property
    def ndim(self) -> int:
        '''张量的维度, 用法同NumPy.
        
        Example
        -------
        >>> from pydynet import Tensor
        >>> Tensor([[2, 2]]).ndim
        2
        '''
        return self.data.ndim

    @property
    def dtype(self) -> npt.DTypeLike:
        '''张量的数据类型, 用法同NumPy.

        Example
        -------
        >>> from pydynet import Tensor
        >>> Tensor([[2, 2]]).dtype
        dtype('int64')
        '''
        return self.data.dtype

    @property
    def size(self) -> int:
        '''张量的元素个数, 用法同NumPy.

        Example
        -------
        >>> from pydynet import Tensor
        >>> Tensor([[1, 1]]).size
        2
        '''
        return self.data.size

    @property
    def strides(self) -> tuple[int]:
        return self.data.strides

    @property
    def T(self) -> Tensor:
        return self.transpose()

    def __repr__(self) -> str:
        return "{}({}, requires_grad={}".format(
            "Tensor",
            self.data,
            self.requires_grad,
        ) + (", device={}".format(self.device)
             if self.device.device != "cpu" else "") + ")"

    def __len__(self) -> int:
        return len(self.data)

    def astype(self, new_type) -> Tensor:
        '''类型转换, 我们不允许可求导节点的类型转换'''
        assert not self.requires_grad
        with self.device:
            return Tensor(
                self.data.astype(new_type),
                new_type,
                copy=None,
                device=self.device,
            )

    def reshape(self, *new_shape) -> Tensor:
        return reshape(self, new_shape)

    def transpose(self, *axes) -> Tensor:
        return transpose(self, axes if len(axes) != 0 else None)

    def swapaxes(self, axis1: int, axis2: int) -> Tensor:
        return swapaxes(self, axis1, axis2)

    def max(self,
            axis: int | tuple | None = None,
            keepdims: bool = False) -> Tensor:
        return max(self, axis, keepdims)

    def min(self,
            axis: int | tuple | None = None,
            keepdims: bool = False) -> Tensor:
        return min(self, axis, keepdims)

    def mean(self,
             axis: int | tuple | None = None,
             keepdims: bool = False) -> Tensor:
        return mean(self, axis, keepdims)

    def sum(self,
            axis: int | tuple | None = None,
            keepdims: bool = False) -> Tensor:
        return sum(self, axis, keepdims)

    def argmax(self,
               axis: int | tuple | None = None,
               keepdims: bool = False) -> Tensor:
        return argmax(self, axis, keepdims)

    def argmin(self,
               axis: int | tuple | None = None,
               keepdims: bool = False) -> Tensor:
        return argmin(self, axis, keepdims)

    def __add__(self, x) -> Tensor:
        return add(self, x)

    def __radd__(self, x) -> Tensor:
        return add(x, self)

    def __sub__(self, x) -> Tensor:
        return sub(self, x)

    def __rsub__(self, x) -> Tensor:
        return sub(x, self)

    def __mul__(self, x) -> Tensor:
        return mul(self, x)

    def __rmul__(self, x) -> Tensor:
        return mul(x, self)

    def __matmul__(self, x) -> Tensor:
        return matmul(self, x)

    def __rmatmul__(self, x) -> Tensor:
        return matmul(x, self)

    def __truediv__(self, x) -> Tensor:
        return div(self, x)

    def __rtruediv__(self, x) -> Tensor:
        return div(x, self)

    def __pow__(self, x) -> Tensor:
        return pow(self, x)

    def __rpow__(self, x) -> Tensor:
        return pow(x, self)

    def __pos__(self) -> Tensor:
        return 1 * self

    def __neg__(self) -> Tensor:
        return -1 * self

    def __abs__(self) -> Tensor:
        return abs(self)

    def __getitem__(self, key) -> Tensor:
        return _get_slice(self, key)

    def _inplace(self, *others: Tensor, func) -> Tensor:
        if self.requires_grad and is_grad_enable():
            raise ValueError(
                "In-place operation is forbidden in node requires grad.")

        others = tuple(other.data if isinstance(other, Tensor) else other
                       for other in others)

        with self.device:
            func(*others)
        return self

    def __setitem__(self, key, value) -> Tensor:
        return self._inplace(key, value, func=self.data.__setitem__)

    def __iadd__(self, other) -> Tensor:
        return self._inplace(other, func=self.data.__iadd__)

    def __isub__(self, other) -> Tensor:
        return self._inplace(other, func=self.data.__isub__)

    def __imul__(self, other) -> Tensor:
        return self._inplace(other, func=self.data.__imul__)

    def __itruediv__(self, other) -> Tensor:
        return self._inplace(other, func=self.data.__itruediv__)

    def __imatmul__(self, other) -> Tensor:
        return self._inplace(other, func=self.data.__imatmul__)

    def _compare(self, other, func) -> Tensor:
        with self.device:
            return Tensor(
                func(self.data,
                     other.data if isinstance(other, Tensor) else other),
                self.xp.bool_, None, self.device, False)

    @no_grad()
    def eq(self, other) -> Tensor:
        return self._compare(other, lambda x, y: x == y)

    @no_grad()
    def ne(self, other) -> Tensor:
        return self._compare(other, lambda x, y: x != y)

    @no_grad()
    def __lt__(self, other) -> Tensor:
        return self._compare(other, lambda x, y: x < y)

    @no_grad()
    def __le__(self, other) -> Tensor:
        return self._compare(other, lambda x, y: x <= y)

    @no_grad()
    def __gt__(self, other) -> Tensor:
        return self._compare(other, lambda x, y: x > y)

    @no_grad()
    def __ge__(self, other) -> Tensor:
        return self._compare(other, lambda x, y: x >= y)

    def backward(self, retain_graph: bool = False):
        '''
        以节点为输出进行反向传播

        Parameters
        ----------
        retain_graph : bool, default=False
            是否保留计算图

        Example
        -------
        >>> from pydynet.tensor import Tensor
        >>> import pydynet.functional as F
        >>> x = Tensor(2., requires_grad=True)
        >>> y = x**2 + x - 1
        >>> y.backward()
        >>> x.grad
        5.
        '''
        if self not in Graph.node_list:
            raise ValueError(
                "Auto-grad is failed because current node is not in graph.")

        if self.size > 1:
            raise ValueError("backward should be called only on a scalar.")

        y_id = Graph.size - Graph.node_list[::-1].index(self) - 1
        with self.device:
            self.grad = self.xp.ones(self.shape, dtype=self.dtype)
            for node in Graph.node_list[y_id::-1]:
                for last in node.last:
                    if last.requires_grad:
                        add_grad = node.grad_fn(last, node.grad)
                        if add_grad.shape != last.shape:
                            # handle broadcast
                            dim1, dim2 = add_grad.ndim, last.ndim
                            add_grad = add_grad.sum(
                                axis=tuple(i for i in range(dim2)
                                           if last.shape[i] == 1),
                                keepdims=True,
                            )
                            if dim1 != dim2:  # dim1 >= dim2 for sure
                                add_grad = add_grad.sum(
                                    tuple(range(dim1 - dim2)))
                        last.grad += add_grad

                # if not retain graph and node is not leaf, free it
                if not retain_graph and not node.is_leaf:
                    Graph._free_node(node)

    def _build_edge(self, node: Tensor):
        node.last.append(self)

    def zero_grad(self):
        '''梯度归零'''
        with self.device:
            self.grad[...] = 0.

    def numpy(self) -> npt.NDArray:
        '''返回Tensor的内部数据, 即NumPy数组(拷贝)'''
        data = self.data
        if self.device != 'cpu':
            data = data.get()
        return data.copy()

    def item(self):
        return self.data.item()

    def to(self, device) -> Tensor:
        device = Device(device)
        if self.device != device:
            self.device = device
            with device:
                self.data = self.xp.asarray(self.data)
                if self.requires_grad:
                    self.grad = self.xp.asarray(self.grad)
        return self

    def cpu(self):
        return self.to("cpu")

    def cuda(self, id: int = 0):
        return self.to(f"cuda:{id}")

    @property
    def xp(self):
        return self.device.xp


class _UnaryOperator(Tensor):
    '''
    一元运算算子的基类, 将一个一元函数抽象成类

    Example
    -------
    >>> class exp(UnaryOperator):
            def forward(self, x: Tensor):
                return np.exp(x.data)
            def grad_fn(self, x: Tensor, grad) -> np.ndarray:
                return self.data * grad
    >>> x = Tensor(1.)
    >>> y = exp(x)
    '''

    def __init__(self, x: Tensor) -> None:
        if not isinstance(x, Tensor):
            x = Tensor(x)
        self.device = x.device
        data = self.forward(x)
        super().__init__(
            data=data,
            device=self.device,
            copy=None,
            dtype=data.dtype,
            requires_grad=is_grad_enable() and x.requires_grad,
        )
        if self.requires_grad:
            x._build_edge(self)

    def forward(self, x: Tensor) -> np.ndarray:
        with self.device:
            return self.forward_(x)

    def forward_(self, x: Tensor) -> np.ndarray:
        '''前向传播函数, 参数为Tensor, 返回的是NumPy数组'''
        raise NotImplementedError

    def grad_fn(self, x: Tensor, grad: npt.NDArray) -> np.ndarray:
        '''
        反向传播函数, 参数为下游节点, 从上游流入该节点梯度。
        注："上游"和"下游"针对的是反向传播, 比如z = f(x, y), x和y是z的下游节点.

        x : Tensor
            下游节点
        grad : ndarray
            上游流入该节点的梯度
        '''
        raise NotImplementedError

    def __repr__(self) -> str:
        return "Tensor({}, op={})".format(self.data, self.__class__.__name__)


class _BinaryOperator(Tensor):
    '''
    二元运算算子的基类, 将一个二元函数抽象成类

    Example
    -------
    >>> add(BinaryOperator):
            def forward(self, x: Tensor, y: Tensor):
                return x.data + y.data
            def grad_fn(self, node: Tensor, grad: npt.NDArray):
                return np.ones(self.shape) * grad
    >>> x = Tensor(1.)
    >>> y = Tensor(2.)
    >>> z = add(x, y)
    '''

    def __init__(self, x: Tensor, y: Tensor) -> None:
        # A strict type-cast: follow the Tensor
        if not isinstance(x, Tensor) and isinstance(y, Tensor):
            x = Tensor(x, dtype=y.dtype, device=y.device)
        elif isinstance(x, Tensor) and not isinstance(y, Tensor):
            y = Tensor(y, dtype=x.dtype, device=x.device)
        elif not (isinstance(x, Tensor) and isinstance(y, Tensor)):
            x, y = Tensor(x), Tensor(y)
        assert x.device == y.device
        # if x.device != 'cpu':
        #     assert x.data.device.id == y.data.device.id
        self.device = x.device
        data = self.forward(x, y)

        super().__init__(
            data=data,
            copy=None,
            device=x.device,
            dtype=data.dtype,
            requires_grad=is_grad_enable()
            and (x.requires_grad or y.requires_grad),
        )
        if self.requires_grad:
            x._build_edge(self)
            y._build_edge(self)

    def forward(self, x: Tensor, y: Tensor) -> np.ndarray:
        with self.device:
            return self.forward_(x, y)

    def forward_(self, x: Tensor, y: Tensor) -> np.ndarray:
        raise NotImplementedError

    def grad_fn(self, x: Tensor, grad: npt.NDArray) -> np.ndarray:
        '''
        反向传播函数, 参数为下游节点, 从上游流入该节点梯度。
        注："上游"和"下游"针对的是反向传播, 比如z = f(x, y), x和y是z的下游节点.

        x : Tensor
            下游节点
        grad : ndarray
            上游流入该节点的梯度
        '''
        raise NotImplementedError

    def __repr__(self) -> str:
        return "Tensor({}, op={})".format(self.data, self.__class__.__name__)


class add(_BinaryOperator):
    '''
    加法算子

    Example
    -------
    >>> x = Tensor(1.)
    >>> y = Tensor(2.)
    >>> z = add(x, y) # 在Tensor类中进行了重载, 所以也可以写成
    >>> z = x + y
    '''

    def forward_(self, x: Tensor, y: Tensor):
        return x.data + y.data

    def grad_fn(self, node: Tensor, grad: npt.NDArray):
        return grad[...]


class sub(_BinaryOperator):
    '''
    减法算子, 在Tensor类中进行重载

    See also
    --------
    add : 加法算子
    '''

    def forward_(self, x: Tensor, y: Tensor):
        return x.data - y.data

    def grad_fn(self, node: Tensor, grad: npt.NDArray):
        if node is self.last[0]:
            return grad[...]
        return -grad


class mul(_BinaryOperator):
    '''
    元素级乘法算子, 在Tensor类中进行重载

    Example
    -------
    >>> x = Tensor([1., 2.])
    >>> y = Tensor([2., 3.])
    >>> z = mul(x, y) # [2, 6]

    See also
    --------
    add : 加法算子
    '''

    def __init__(self, x: Tensor, y: Tensor) -> None:
        super().__init__(x, y)

    def forward_(self, x: Tensor, y: Tensor):
        return x.data * y.data

    def grad_fn(self, node: Tensor, grad: npt.NDArray):
        if node is self.last[0]:
            return self.last[1].data * grad
        return self.last[0].data * grad


class div(_BinaryOperator):
    '''
    除法算子, 在Tensor类中进行重载

    See also
    --------
    add : 加法算子
    '''

    def __init__(self, x: Tensor, y: Tensor) -> None:
        super().__init__(x, y)

    def forward_(self, x: Tensor, y: Tensor):
        return x.data / y.data

    def grad_fn(self, node: Tensor, grad: npt.NDArray):
        temp = grad / self.last[1].data
        if node is self.last[0]:
            return temp
        return -self.data * temp


class pow(_BinaryOperator):
    '''
    幂运算算子, 在Tensor类中进行重载

    See also
    --------
    add : 加法算子
    '''

    def __init__(self, x: Tensor, y: Tensor) -> None:
        super().__init__(x, y)

    def forward_(self, x: Tensor, y: Tensor):
        return x.data**y.data

    def grad_fn(self, node: Tensor, grad: npt.NDArray) -> np.ndarray:
        if node is self.last[0]:
            return (self.data * self.last[1].data / node.data) * grad
        else:
            return self.data * self.xp.log(self.last[0].data) * grad


class matmul(_BinaryOperator):
    '''
    矩阵乘法算子, 在Tensor类中进行重载, 张量的矩阵乘法遵从NumPy Matmul的规则.

    参考 : https://xingcy.net/2022/04/26/broadcast/

    See also
    --------
    add : 加法算子
    '''

    def __init__(self, x: Tensor, y: Tensor) -> None:
        super().__init__(x, y)

    def forward_(self, x: Tensor, y: Tensor) -> np.ndarray:
        self.expand_a, self.expand_b = x.ndim < 2, y.ndim < 2
        return x.data @ y.data

    def grad_fn(self, node: Tensor, grad: npt.NDArray) -> np.ndarray:
        # regularization for input shape
        if self.expand_a:
            grad = self.xp.expand_dims(grad, 0)

        if self.expand_b:
            grad = self.xp.expand_dims(grad, -1)

        if node is self.last[0]:
            grad1 = grad @ (self.xp.atleast_2d(self.last[1].data) if self.
                            expand_b else self.last[1].data.swapaxes(-1, -2))
            return grad1[0] if self.expand_a else grad1
        else:
            grad2 = (self.xp.atleast_2d(self.last[0].data) if self.expand_a
                     else self.last[0].data).swapaxes(-1, -2) @ grad
            return grad2[..., 0] if self.expand_b else grad2


class abs(_UnaryOperator):
    '''
    绝对值算子, 在Tensor类中进行重载

    See also
    --------
    add : 加法算子
    '''

    def forward_(self, x: Tensor) -> np.ndarray:
        return self.xp.abs(x.data)

    def grad_fn(self, x: Tensor, grad: npt.NDArray) -> np.ndarray:
        return grad * self.xp.sign(x)


class _ReduceOperator(_UnaryOperator):

    def __init__(self, x: Tensor, axis=None, keepdims=False, func: str = ...):
        self.axis = axis
        self.keepdims = keepdims
        self._scalar_or_keeps = self.axis is None or self.keepdims
        self.reduce_func = getattr(x.xp, func)
        super().__init__(x)

    def forward_(self, x):
        return self.reduce_func(x.data, axis=self.axis, keepdims=self.keepdims)

    def grad_fn(self, x, grad):
        if not (self.axis is None or self.keepdims):
            grad = self.xp.expand_dims(grad, axis=self.axis)

        return self._reduce_grad_fn(x.data, grad)

    def _reduce_grad_fn(self, x_data, grad):
        raise NotImplementedError


class sum(_ReduceOperator):

    def __init__(self, x, axis=None, keepdims=False):
        super().__init__(x, axis, keepdims, 'sum')

    def _reduce_grad_fn(self, x_data: npt.NDArray, grad):
        return self.xp.broadcast_to(grad, x_data.shape)


class mean(_ReduceOperator):

    def __init__(self, x, axis=None, keepdims=False):
        super().__init__(x, axis, keepdims, 'mean')

    def _reduce_grad_fn(self, x_data: npt.NDArray, grad):
        return self.xp.broadcast_to(grad,
                                    x_data.shape) * self.size / x_data.size


class max(_ReduceOperator):

    def __init__(self, x, axis=None, keepdims=False):
        super().__init__(x, axis, keepdims, 'max')

    def grad_fn(self, x: Tensor, grad: npt.NDArray) -> np.ndarray:
        if self._scalar_or_keeps:
            full_dim_y = self.data
        else:
            full_dim_y = self.xp.expand_dims(self.data, axis=self.axis)
            grad = self.xp.expand_dims(grad, axis=self.axis)
        return (full_dim_y == x.data) * grad


class min(_ReduceOperator):

    def __init__(self, x, axis=None, keepdims=False):
        super().__init__(x, axis, keepdims, 'min')

    def grad_fn(self, x: Tensor, grad: npt.NDArray) -> np.ndarray:
        if self._scalar_or_keeps:
            full_dim_y = self.data
        else:
            full_dim_y = self.xp.expand_dims(self.data, axis=self.axis)
            grad = self.xp.expand_dims(grad, axis=self.axis)
        return (full_dim_y == x.data) * grad


class argmax(_ReduceOperator):

    def __init__(self, x, axis=None, keepdims=False):
        super().__init__(x, axis, keepdims, 'argmax')


class argmin(_ReduceOperator):

    def __init__(self, x, axis=None, keepdims=False):
        super().__init__(x, axis, keepdims, 'argmin')


class exp(_UnaryOperator):
    '''指数运算
    
    Example
    -------
    >>> x = Tensor(1.)
    >>> y = exp(x)
    '''

    def forward_(self, x: Tensor):
        return self.xp.exp(x.data)

    def grad_fn(self, x: Tensor, grad: npt.NDArray) -> np.ndarray:
        return self.data * grad


class log(_UnaryOperator):
    '''对数运算
    
    Example
    -------
    >>> x = Tensor(1.)
    >>> y = log(x)
    '''

    def forward_(self, x: Tensor):
        return self.xp.log(x.data)

    def grad_fn(self, x: Tensor, grad: npt.NDArray) -> np.ndarray:
        return grad / x.data


class maximum(_BinaryOperator):

    def forward_(self, x: Tensor, y: Tensor) -> np.ndarray:
        return self.xp.maximum(x.data, y.data)

    def grad_fn(self, x: Tensor, grad) -> np.ndarray:
        return (self.data == x.data) * grad


class minimum(_BinaryOperator):

    def forward_(self, x: Tensor, y: Tensor) -> np.ndarray:
        return self.xp.minimum(x.data, y.data)

    def grad_fn(self, x: Tensor, grad) -> np.ndarray:
        return (self.data == x) * grad


class sign(_UnaryOperator):

    def forward_(self, x: Tensor) -> np.ndarray:
        return self.xp.sign(x.data)

    def grad_fn(self, x, grad):
        return self.xp.zeros(self.shape, dtype=self.dtype)


# 非计算函数
class reshape(_UnaryOperator):
    '''
    张量形状变换算子, 在Tensor中进行重载

    Parameters
    ----------
    new_shape : tuple
        变换后的形状, 用法同NumPy
    '''

    def __init__(self, x: Tensor, new_shape: tuple) -> None:
        self.new_shape = new_shape
        super().__init__(x)

    def forward_(self, x: Tensor) -> np.ndarray:
        return x.data.reshape(self.new_shape)

    def grad_fn(self, x: Tensor, grad: npt.NDArray) -> np.ndarray:
        return grad.reshape(x.shape)


class transpose(_UnaryOperator):
    '''
    张量转置算子, 在Tensor中进行重载(Tensor.T和Tensor.transpose)

    Parameters
    ----------
    axes : tuple
        转置的轴变换, 用法同NumPy
    '''

    def __init__(self, x: Tensor, axes: tuple = None) -> None:
        self.axes = axes
        super().__init__(x)

    def forward_(self, x: Tensor) -> np.ndarray:
        return x.data.transpose(self.axes)

    def grad_fn(self, x: Tensor, grad: npt.NDArray) -> np.ndarray:
        if self.axes is None:
            return grad.transpose()
        return grad.transpose(tuple(np.argsort(self.axes)))


class swapaxes(_UnaryOperator):
    '''
    张量交换轴算子

    Parameters
    ----------
    axis1 : int
        第一个axis;
    axis2 : int
        第二个axis.
    '''

    def __init__(self, x: Tensor, axis1: int, axis2: int) -> None:
        self.axis1 = axis1
        self.axis2 = axis2
        super().__init__(x)

    def forward_(self, x: Tensor) -> np.ndarray:
        return x.data.swapaxes(self.axis1, self.axis2)

    def grad_fn(self, x: Tensor, grad: npt.NDArray) -> np.ndarray:
        return grad.swapaxes(self.axis1, self.axis2)


class _get_slice(_UnaryOperator):
    '''
    切片算子, 为Tensor类提供索引和切片接口

    Example
    -------
    >>> x = Tensor(
            np.arange(12).reshape(3, 4).astype(float),
            requires_grad=True,
        )
    >>> y = x[:2, :2].sum()
    >>> y.backward()
    >>> x.grad 
    [[1. 1. 0. 0.]
     [1. 1. 0. 0.]
     [0. 0. 0. 0.]]
    '''

    def __init__(self, x: Tensor, key) -> None:
        if isinstance(key, tuple):
            new_key = []
            for k in key:
                new_key.append(k if not isinstance(k, Tensor) else k.data)
            self.key = tuple(new_key)
        elif isinstance(key, Tensor):
            self.key = key.data
        else:
            self.key = key
        super().__init__(x)

    def forward_(self, x: Tensor) -> np.ndarray:
        return x.data[self.key]

    def grad_fn(self, x: Tensor, grad: npt.NDArray) -> np.ndarray:
        full_grad = self.xp.zeros(x.shape, dtype=x.dtype)
        full_grad[self.key] = grad
        return full_grad


class concat(Tensor):
    '''对多个张量进行连接, 用法类似于`numpy.concatenate`
    
    Parameters
    ----------
    tensors : 
        待连接的张量：
    axis : default=0
        连接轴, 默认是沿着第一个轴拼接.
    '''

    def __init__(self, tensors: list[Tensor], axis=0) -> None:
        requires_grad = False
        self.tensors = tensors
        self.axis = axis
        self.indices = [0]

        for i in range(len(self.tensors)):
            assert isinstance(
                tensors[i],
                Tensor), "Concatenate elements in 'tensors' must be 'Tensor'"
            if i == 0:
                device = tensors[i].device
            else:
                assert tensors[i].device == device
            requires_grad = requires_grad or self.tensors[i].requires_grad
            self.indices.append(self.indices[-1] +
                                self.tensors[i].shape[self.axis])
        self.device = device
        super().__init__(
            self.forward(),
            requires_grad=requires_grad and is_grad_enable(),
            copy=False,
            device=self.device,
        )
        if self.requires_grad:
            for i in range(len(self.tensors)):
                self.tensors[i]._build_edge(self)

    def forward(self):
        with self.device:
            return self.xp.concatenate([t.data for t in self.tensors],
                                       axis=self.axis)

    def grad_fn(self, x, grad: npt.NDArray):
        x_id = self.tensors.index(x)
        start = self.indices[x_id]
        end = self.indices[x_id + 1]
        slc = [slice(None)] * grad.ndim
        slc[self.axis] = slice(start, end)
        return grad[tuple(slc)]


class sigmoid(_UnaryOperator):
    '''Sigmoid运算, 我们前向传播避免了溢出问题'''

    def forward_(self, x: Tensor) -> np.ndarray:
        sigmoid = self.xp.zeros(x.shape, dtype=x.dtype)
        sigmoid[x.data > 0] = 1 / (1 + self.xp.exp(-x.data[x.data > 0]))
        sigmoid[x.data <= 0] = 1 - 1 / (1 + self.xp.exp(x.data[x.data <= 0]))
        return sigmoid

    def grad_fn(self, x: Tensor, grad: npt.NDArray) -> np.ndarray:
        return self.data * (1 - self.data) * grad


class tanh(_UnaryOperator):
    '''Tanh运算, 我们前向传播避免了溢出问题'''

    def forward_(self, x: Tensor) -> np.ndarray:
        tanh = self.xp.zeros(x.shape, dtype=x.dtype)
        tanh[x.data > 0] = 2 / (1 + self.xp.exp(-2 * x.data[x.data > 0])) - 1
        tanh[x.data <= 0] = 1 - 2 / (1 + self.xp.exp(2 * x.data[x.data <= 0]))
        return tanh

    def grad_fn(self, x: Tensor, grad: npt.NDArray) -> np.ndarray:
        return (1 - self.data**2) * grad



import numpy as np
import warnings

try:
    import cupy as cp
    cuda_available: bool = True
except ModuleNotFoundError:
    warnings.warn(
        "Cupy is not installed. You can install it with:\n"
        "  pip install cupy-cuda12x  # or appropriate version for your CUDA",
        category=UserWarning)
    cuda_available: bool = False
    cp = object()


def is_available() -> bool:
    return cuda_available


def device_count() -> int:
    if is_available():
        return cp.cuda.runtime.getDeviceCount()
    else:
        return 0


def current_device() -> int:
    return cp.cuda.runtime.getDevice()


def set_device(device: int) -> None:
    return cp.cuda.runtime.setDevice(device)


class Device:

    def __init__(self, device=None) -> None:
        if isinstance(device, str):
            if device == "cpu":
                self.device = "cpu"
            elif device[:4] == "cuda":
                self.device = "cuda"
                if len(device) == 4:
                    device += ':0'

                cuda_id = device.split(':')[-1]
                if not cuda_id.isdigit():
                    raise ValueError(f'Wrong cuda id \"{cuda_id}\"!')

                self.device_id = int(cuda_id)
            else:
                raise ValueError(f"Unknown device \"{device}\"!")

        elif isinstance(device, int):
            self.device = "cuda"
            self.device_id = device

        elif device is None:
            self.device = "cpu"

        elif isinstance(device, Device):
            self.device = device.device
            if self.device != "cpu":
                self.device_id = device.device_id

        if self.device == "cuda":
            if not is_available():
                raise RuntimeError(
                    "Cuda device is not supported on this system.")
            self.device = cp.cuda.Device(self.device_id)
        assert self.device == "cpu" or is_available()

    def __repr__(self) -> str:
        if self.device == "cpu":
            return "Device(type='cpu')"
        else:
            return "Device(type='cuda', index={})".format(self.device_id)

    def __eq__(self, device) -> bool:
        if not isinstance(device, Device):
            device = Device(device)
        if self.device == "cpu":
            return device.device == "cpu"
        else:
            if device.device == "cpu":
                return False
            return self.device == device.device

    @property
    def xp(self):
        return np if self.device == "cpu" else cp

    def __enter__(self):
        if self.device != "cpu" and self.device_id != current_device():
            return self.device.__enter__()

    def __exit__(self, type, value, trace):
        if self.device != "cpu" and self.device_id != current_device():
            return self.device.__exit__(type, value, trace)



from .activation import Sigmoid, Tanh, ReLU, LeakyReLU, Softmax
from .norm import BatchNorm1d, BatchNorm2d, LayerNorm, RMSNorm
from .conv import Conv1d, Conv2d
from .pool import MaxPool1d, MaxPool2d, AvgPool1d, AvgPool2d
from .dropout import Dropout
from .linear import Linear, Embedding
from .loss import MSELoss, NLLLoss, CrossEntropyLoss
from .module import Module, Sequential, ModuleList
from .rnn import RNN, LSTM, GRU, RNNCell, LSTMCell, GRUCell

__all__ = [
    "Sigmoid",
    "Tanh",
    "ReLU",
    "LeakyReLU",
    "Softmax",
    "BatchNorm1d",
    "BatchNorm2d",
    "LayerNorm",
    "RMSNorm",
    "Conv1d",
    "Conv2d",
    "MaxPool1d",
    "MaxPool2d",
    "AvgPool1d",
    "AvgPool2d",
    "Dropout",
    "Linear",
    "Embedding",
    "MSELoss",
    "NLLLoss",
    "CrossEntropyLoss",
    "Module",
    "Sequential",
    "ModuleList",
    "RNN",
    "LSTM",
    "GRU",
    "RNNCell",
    "LSTMCell",
    "GRUCell",
]



from ..core import Tensor
from ..autograd import no_grad
import math


def calculate_gain(nonlinearity: str, param: float = None) -> float:
    return {
        "linear": 1,
        "conv1d": 1,
        "conv2d": 1,
        "sigmoid": 1,
        "tanh": 5 / 3,
        "relu": math.sqrt(2.),
        "leaky_relu":
        math.sqrt(2. / (1 + (param if param != None else 0.01)**2))
    }[nonlinearity]


def _calculate_fan(tensor: Tensor):
    assert tensor.ndim >= 2
    fan_in, fan_out = tensor.shape[:2]
    if tensor.ndim > 2:
        receptive_field_size = math.prod(tensor.shape[2:])
        fan_in *= receptive_field_size
        fan_out *= receptive_field_size
    return fan_in, fan_out


@no_grad()
def uniform_(tensor: Tensor, a=0., b=1.) -> Tensor:
    tensor.data[...] = tensor.xp.random.uniform(a, b, tensor.shape)
    return tensor


@no_grad()
def normal_(tensor: Tensor, mean=0., std=1.) -> Tensor:
    tensor.data[...] = tensor.xp.random.normal(mean, std, size=tensor.shape)
    return tensor


@no_grad()
def constant_(tensor: Tensor, val: float) -> Tensor:
    tensor.data[...] = val
    return tensor


def ones_(tensor: Tensor) -> Tensor:
    return constant_(tensor, 1.)


def zeros_(tensor: Tensor) -> Tensor:
    return constant_(tensor, 0.)


def xavier_uniform_(tensor: Tensor, gain: float = 1.) -> Tensor:
    fan_in, fan_out = _calculate_fan(tensor)
    bound = gain * math.sqrt(6. / (fan_in + fan_out))
    return uniform_(tensor, -bound, bound)


def xavier_normal_(tensor: Tensor, gain: float = 1.) -> Tensor:
    fan_in, fan_out = _calculate_fan(tensor)
    std = gain * math.sqrt(2 / (fan_in + fan_out))
    return normal_(tensor, std=std)


def kaiming_uniform_(tensor: Tensor,
                     a: float = 0.,
                     mode='fan_in',
                     nonlinearity='relu') -> Tensor:
    fan_in, fan_out = _calculate_fan(tensor)
    fan = {
        "fan_in": fan_in,
        "fan_out": fan_out,
    }[mode]
    gain = calculate_gain(nonlinearity, a)
    bound = gain * math.sqrt(3. / fan)
    return uniform_(tensor, -bound, bound)


def kaiming_normal_(tensor: Tensor,
                    a: float = 0.,
                    mode='fan_in',
                    nonlinearity='relu'):
    fan_in, fan_out = _calculate_fan(tensor)
    fan = {
        "fan_in": fan_in,
        "fan_out": fan_out,
    }[mode]
    gain = calculate_gain(nonlinearity, a)
    std = gain / math.sqrt(fan)
    return normal_(tensor, std=std)



'''优化器类，我们目前实现了\n
- SGD with momentum and Nestrov;\n
- Adagrad;\n
- Adadelta;\n
- Adam.\n

Reference
---------
论文: https://arxiv.org/abs/1609.04747;\n
博客: https://xingcy.net/2021/08/20/gd/.
'''

from math import sqrt
from ..core import Tensor


class Optimizer:
    '''优化器基类'''

    def __init__(self, params: list[Tensor]) -> None:
        self.params: list[Tensor] = list(params)

    def step(self):
        raise NotImplementedError

    def zero_grad(self):
        '''针对self.params梯度清零.'''
        for param in self.params:
            param.zero_grad()


class SGD(Optimizer):
    '''带动量的梯度下降

    Parameters
    ----------
    params : list[Parameter]
        待优化参数;
    lr : float
        学习率;
    momentum : float
        动量系数;
    weight_decay : float, default=0.
        权重衰减系数.
    nesterov : bool, defallt=True.
        是否采用Nesterov加速.
    '''

    def __init__(
        self,
        params: list[Tensor],
        lr: float,
        momentum: float = .5,
        weight_decay: float = 0.,
        nesterov=True,
    ) -> None:
        super().__init__(params)
        self.lr = lr
        self.momentum = momentum
        self.weight_decay = weight_decay
        self.nesterov = nesterov
        self.v = [
            param.xp.zeros(param.shape, dtype=param.dtype)
            for param in self.params
        ]

    def step(self):
        for i in range(len(self.params)):
            with self.params[i].device:
                grad = self.params[i].grad + self.weight_decay * self.params[i].data
                self.v[i] *= self.momentum
                self.v[i] += self.lr * grad
                self.params[i].data -= self.v[i]
                if self.nesterov:
                    self.params[i].data -= self.lr * grad


class Adagrad(Optimizer):
    '''Adaptive Gradient Descent
    
    Parameters
    ----------
    params : list[Parameter]
        待优化参数;
    lr : float, default=1e-2.
        学习率;
    weight_decay : float, default=0.
        权重衰减系数.
    eps : float, default=1e-10
        epsilon.
    '''

    def __init__(
        self,
        params: list[Tensor],
        lr: float = 1e-2,
        weight_decay: float = 0,
        eps: float = 1e-10,
    ) -> None:
        super().__init__(params)
        self.lr = lr
        self.weight_decay = weight_decay
        self.eps = eps
        self.G = [
            param.xp.zeros(param.shape, dtype=param.dtype)
            for param in self.params
        ]

    def step(self):
        for i in range(len(self.params)):
            with self.params[i].device:
                grad = self.params[i].grad + self.weight_decay * self.params[i].data
                self.G[i] += grad**2
                self.params[i].data -= self.lr * grad / (self.eps + self.G[i])**0.5


class Adadelta(Optimizer):
    '''
    Adadelta优化器
    
    params : list[Parameter]
        待优化参数;
    lr : float, default=1e-2.
        学习率;
    rho :float, default=
    weight_decay : float, default=0.
        权重衰减系数.
    eps : float, default=1e-10
        epsilon.
    '''

    def __init__(
        self,
        params: list[Tensor],
        lr: float = 1.0,
        rho: float = 0.9,
        weight_decay: float = 0,
        eps: float = 1e-6,
    ) -> None:
        super().__init__(params)
        self.lr = lr
        self.rho = rho
        self.eps = eps
        self.eps = eps
        self.weight_decay = weight_decay
        self.G = [
            param.xp.zeros(param.shape, dtype=param.dtype)
            for param in self.params
        ]

    def step(self):
        for i in range(len(self.params)):
            with self.params[i].device:
                grad = self.params[i].grad + self.weight_decay * self.params[i].data

                self.G[i] = self.rho * self.G[i] + (1 - self.rho) * grad**2
                self.params[i].data -= self.lr * grad / (self.G[i] + self.eps)**0.5


class Adam(Optimizer):

    def __init__(
        self,
        params: list[Tensor],
        lr: float = 1e-3,
        betas: tuple[float] = (0.9, 0.999),
        eps: float = 1e-8,
        weight_decay: float = 0,
    ) -> None:
        super().__init__(params)
        self.lr = lr
        self.beta1, self.beta2 = betas
        self.eps = eps
        self.weight_decay = weight_decay
        self.m = [
            param.xp.zeros(param.shape, dtype=param.dtype)
            for param in self.params
        ]
        self.v = [
            param.xp.zeros(param.shape, dtype=param.dtype)
            for param in self.params
        ]
        self.t = 1

    def step(self):
        for i in range(len(self.params)):
            with self.params[i].device:
                grad = self.params[i].grad + self.weight_decay * self.params[i].data
                self.m[i] *= self.beta1
                self.m[i] += (1 - self.beta1) * grad
                self.v[i] *= self.beta2
                self.v[i] += (1 - self.beta2) * grad**2
                a_t = sqrt(1 - self.beta2**self.t) / (1 - self.beta1**self.t)
                self.params[i].data -= self.lr * a_t * self.m[i] / (
                    self.v[i]**0.5 + self.eps)
        self.t += 1



import setuptools

setuptools.setup(
    name='pydynet',
    version='1.1',
    description=
    'PyDyNet: Neuron Network (MLP, CNN, RNN, Transformer, ...) implementation using Numpy with Autodiff',
    author="Cun-Yuan Xing",
    author_email="xingcy@lamda.nju.edu.cn",
    maintainer="Cun-Yuan Xing",
    maintainer_email="xingcy@lamda.nju.edu.cn",
    packages=[
        'pydynet', 'pydynet/optim', 'pydynet/nn', 'pydynet/nn/modules',
        'pydynet/core'
    ],
    license='MIT License',
    install_requires=['numpy>=2.0.0'],
    long_description=open('README.md', encoding='utf-8').read(),
    long_description_content_type="text/markdown",
    url='https://github.com/WeltXing/PyDyNet',
)



from .module import Module
from .. import init
from .. import functional as F
from ..parameter import Parameter
from ...special import empty, zeros
from ... import core
from ...cuda import Device

from typing import Literal, Optional, Tuple, List
import math


class RNNCell(Module):

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        bias: bool = True,
        nonlinearity: Literal['tanh', 'relu'] = 'tanh',
        device=None,
        dtype=None,
    ) -> None:
        super().__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.kwargs = {"device": Device(device), "dtype": dtype}
        self.nonlinearity = nonlinearity
        self.fn = {'tanh': F.tanh, 'relu': F.relu}[nonlinearity]

        self.Wx = Parameter(empty((input_size, hidden_size), **self.kwargs))
        self.Wh = Parameter(empty((hidden_size, hidden_size), **self.kwargs))
        if bias:
            self.bias = Parameter(empty(self.hidden_size, **self.kwargs))
        self.has_bias = bias
        self.reset_paramters()

    def forward(self, x, h=None):
        if h is None:
            h = self.init_hidden(x)
        else:
            assert (x.ndim == 1 and h.shape == (self.hidden_size, )) or (
                x.ndim == 2 and h.shape
                == (x.shape[0], self.hidden_size)), "Wrong hidden state input!"

        lin = x @ self.Wx + h @ self.Wh
        if self.has_bias:
            lin = lin + self.bias
        return self.fn(lin)

    def reset_paramters(self):
        bound = math.sqrt(1 / self.hidden_size)
        init.uniform_(self.Wx, -bound, bound)
        init.uniform_(self.Wh, -bound, bound)
        if self.has_bias:
            init.uniform_(self.bias, -bound, bound)

    def init_hidden(self, x):
        assert x.ndim in {1, 2}
        if x.ndim == 1:
            return zeros(self.hidden_size, **self.kwargs)
        else:
            batch_size = x.shape[0]
            return zeros((batch_size, self.hidden_size), **self.kwargs)

    def __repr__(self) -> str:
        return "{}({}, {}, bias={}, nonlinearity={})".format(
            self.__class__.__name__,
            self.input_size,
            self.hidden_size,
            self.has_bias,
            self.nonlinearity,
        )

    def move(self, device):
        self.kwargs['device'] = device
        return super().move(device)


class RNN(Module):

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_layers: int = 1,
        nonlinearity: Literal['tanh', 'relu'] = 'tanh',
        bias: bool = True,
        batch_first: bool = False,
        bidirectional: bool = False,
        device=None,
        dtype=None,
    ) -> None:
        super().__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.nonlinearity = nonlinearity
        self.has_bias = bias
        self.batch_first = batch_first
        self.bidirectional = bidirectional
        self.kwargs = {"device": Device(device), "dtype": dtype}

        assert num_layers > 0
        size_list = [input_size] + [hidden_size] * (num_layers - 1)
        self.RNNCells: List[RNNCell] = []
        for i in range(num_layers):
            cell = RNNCell(
                size_list[i],
                hidden_size,
                bias,
                nonlinearity,
                **self.kwargs,
            )
            setattr(self, 'rnn_{}'.format(i), cell)
            self.RNNCells.append(cell)
        if self.bidirectional:
            self.rRNNCells: List[RNNCell] = []
            for i in range(num_layers):
                cell = RNNCell(
                    size_list[i],
                    hidden_size,
                    bias,
                    nonlinearity,
                    **self.kwargs,
                )
                setattr(self, 'rrnn_{}'.format(i), cell)
                self.rRNNCells.append(cell)

    def forward(self, x, h=None):
        if self.batch_first and x.ndim == 3:
            x = x.swapaxes(0, 1)

        if h is None:
            h = self.init_hidden(x)
        else:
            d = 2 if self.bidirectional else 1
            assert (x.ndim == 2
                    and h.shape == (d * self.num_layers, self.hidden_size)
                    ) or (x.ndim == 3 and h.shape
                          == (d * self.num_layers, x.shape[1],
                              self.hidden_size)), "Wrong hidden state input!"

        if self.num_layers == 1 and not self.bidirectional:
            h_list = self.cell_forward(self.RNNCells[0], x, h[0])
            output = core.concat(h_list)
            hn = h_list[-1]

        elif self.num_layers == 1 and self.bidirectional:
            h_list = self.cell_forward(self.RNNCells[0], x, h[0])
            hr_list = self.cell_forward(self.rRNNCells[0], x[::-1], h[1])
            output = core.concat(
                [core.concat(h_list),
                 core.concat(hr_list[::-1])],
                axis=-1,
            )
            hn = core.concat([h_list[-1], hr_list[-1]])

        elif self.num_layers > 1 and not self.bidirectional:
            hn_list = []
            for i in range(self.num_layers):
                h_list = self.cell_forward(
                    self.RNNCells[i],
                    x if i == 0 else core.concat(h_list),
                    h[i],
                )
                hn_list.append(h_list[-1])
            output = core.concat(h_list)
            hn = core.concat(hn_list)

        else:
            hn_list = []
            hrn_list = []
            for i in range(self.num_layers):
                h_list = self.cell_forward(
                    self.RNNCells[i],
                    x if i == 0 else core.concat(h_list),
                    h[i],
                )
                hr_list = self.cell_forward(
                    self.rRNNCells[i],
                    x[::-1] if i == 0 else core.concat(hr_list),
                    h[i + self.num_layers],
                )
                hn_list.append(h_list[-1])
                hrn_list.append(hr_list[-1])
            output = core.concat(
                [core.concat(h_list),
                 core.concat(hr_list[::-1])], axis=-1)
            hn = core.concat(hn_list + hrn_list)

        if self.batch_first and x.ndim == 3:
            output = output.swapaxes(0, 1)
            hn = hn.swapaxes(0, 1)
        return output, hn

    def reset_parameters(self):
        for i in range(self.num_layers):
            self.RNNCells[i].reset_paramters()
        if self.bidirectional:
            for i in range(self.num_layers):
                self.rRNNCells[i].reset_paramters()

    def init_hidden(self, x):
        assert x.ndim in {2, 3}
        d = 2 if self.bidirectional else 1
        if x.ndim == 2:
            return zeros(
                (d * self.num_layers, self.hidden_size),
                **self.kwargs,
            )
        else:
            batch_size = x.shape[1]
            return zeros(
                (d * self.num_layers, batch_size, self.hidden_size),
                **self.kwargs,
            )

    def cell_forward(self, cell: RNNCell, x, h):
        seq_len = x.shape[0]
        h_list = []
        for i in range(seq_len):
            h = cell(x[i], h)
            h_list.append(core.unsqueeze(h, axis=0))
        return h_list

    def __repr__(self) -> str:
        return "{}({}, {}, num_layers={}, nonlinearity={}, bias={}, batch_first={}, bidirectional={})".format(
            self.__class__.__name__,
            self.input_size,
            self.hidden_size,
            self.num_layers,
            self.nonlinearity,
            self.has_bias,
            self.batch_first,
            self.bidirectional,
        )

    def move(self, device):
        self.kwargs['device'] = device
        return super().move(device)


class LSTMCell(Module):

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        bias: bool = True,
        device=None,
        dtype=None,
    ) -> None:
        super().__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.kwargs = {"device": Device(device), "dtype": dtype}

        self.Wx = Parameter(empty((input_size, 4 * hidden_size),
                                  **self.kwargs))
        self.Wh = Parameter(
            empty((hidden_size, 4 * hidden_size), **self.kwargs))
        if bias:
            self.bias = Parameter(empty(4 * self.hidden_size, **self.kwargs))
        self.has_bias = bias
        self.reset_paramters()

    def forward(self, x, hx: Optional[Tuple] = None):
        if hx is None:
            h = self.init_hidden(x)
            c = self.init_hidden(x)
        else:
            h, c = hx
            assert (x.ndim == 1 and h.shape == (self.hidden_size, )) or (
                x.ndim == 2 and h.shape
                == (x.shape[0], self.hidden_size)), "Wrong hidden state input!"
            assert (x.ndim == 1 and c.shape == (self.hidden_size, )) or (
                x.ndim == 2 and c.shape
                == (x.shape[0], self.hidden_size)), "Wrong cell state input!"
        lin = x @ self.Wx + h @ self.Wh
        if self.has_bias:
            lin = lin + self.bias
        fio, g = core.hsplit(lin, [3 * self.hidden_size])
        sig_fio, tanh_g = F.sigmoid(fio), F.tanh(g)
        f, i, o = core.hsplit(sig_fio, 3)
        c = f * c + i * tanh_g
        h = o * F.tanh(c)
        return h, c

    def init_hidden(self, x):
        assert x.ndim in {1, 2}
        if x.ndim == 1:
            return zeros(self.hidden_size, **self.kwargs)
        else:
            batch_size = x.shape[0]
            return zeros((batch_size, self.hidden_size), **self.kwargs)

    def reset_paramters(self):
        bound = math.sqrt(1 / self.hidden_size)
        init.uniform_(self.Wx, -bound, bound)
        init.uniform_(self.Wh, -bound, bound)
        if self.has_bias:
            init.uniform_(self.bias, -bound, bound)

    def __repr__(self) -> str:
        return "{}({}, {}, bias={})".format(
            self.__class__.__name__,
            self.input_size,
            self.hidden_size,
            self.has_bias,
        )

    def move(self, device):
        self.kwargs['device'] = device
        return super().move(device)


class LSTM(Module):

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_layers: int = 1,
        bias: bool = True,
        batch_first: bool = False,
        bidirectional: bool = False,
        device=None,
        dtype=None,
    ) -> None:
        super().__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.has_bias = bias
        self.batch_first = batch_first
        self.bidirectional = bidirectional
        self.kwargs = {"device": Device(device), "dtype": dtype}

        assert num_layers > 0
        size_list = [input_size] + [hidden_size] * (num_layers - 1)
        self.LSTMCells: List[LSTMCell] = []
        for i in range(num_layers):
            cell = LSTMCell(
                size_list[i],
                hidden_size,
                bias,
                **self.kwargs,
            )
            setattr(self, 'lstm_{}'.format(i), cell)
            self.LSTMCells.append(cell)
        if self.bidirectional:
            self.rLSTMCells: List[LSTMCell] = []
            for i in range(num_layers):
                cell = LSTMCell(
                    size_list[i],
                    hidden_size,
                    bias,
                    **self.kwargs,
                )
                setattr(self, 'rlstm_{}'.format(i), cell)
                self.rLSTMCells.append(cell)

    def forward(self, x, hx: Optional[Tuple] = None):
        if self.batch_first and x.ndim == 3:
            x = x.swapaxes(0, 1)

        if hx is None:
            h = self.init_hidden(x)
            c = self.init_hidden(x)
        else:
            d = 2 if self.bidirectional else 1
            h, c = hx
            assert (x.ndim == 2
                    and h.shape == (d * self.num_layers, self.hidden_size)
                    ) or (x.ndim == 3 and h.shape
                          == (d * self.num_layers, x.shape[1],
                              self.hidden_size)), "Wrong hidden state input!"
            assert (x.ndim == 2
                    and c.shape == (d * self.num_layers, self.hidden_size)
                    ) or (x.ndim == 3 and c.shape
                          == (d * self.num_layers, x.shape[1],
                              self.hidden_size)), "Wrong cell state input!"

        if self.num_layers == 1 and not self.bidirectional:
            h_list, c_list = self.cell_forward(
                self.LSTMCells[0],
                x,
                (h[0], c[0]),
            )
            output = core.concat(h_list)
            hn = h_list[-1]
            cn = c_list[-1]
        elif self.num_layers == 1 and self.bidirectional:
            h_list, c_list = self.cell_forward(
                self.LSTMCells[0],
                x,
                (h[0], c[0]),
            )
            hr_list, cr_list = self.cell_forward(
                self.rLSTMCells[0],
                x[::-1],
                (h[1], c[1]),
            )
            output = core.concat(
                [core.concat(h_list),
                 core.concat(hr_list[::-1])], axis=-1)
            hn = core.concat([h_list[-1], hr_list[-1]])
            cn = core.concat([c_list[-1], cr_list[-1]])
        elif self.num_layers > 1 and not self.bidirectional:
            hn_list, cn_list = [], []
            for i in range(self.num_layers):
                h_list, c_list = self.cell_forward(
                    self.LSTMCells[i],
                    x if i == 0 else core.concat(h_list),
                    (h[i], c[i]),
                )
                hn_list.append(h_list[-1])
                cn_list.append(c_list[-1])
            output = core.concat(h_list)
            hn = core.concat(hn_list)
            cn = core.concat(cn_list)
        else:
            hn_list, hrn_list = [], []
            cn_list, crn_list = [], []
            for i in range(self.num_layers):
                h_list, c_list = self.cell_forward(
                    self.LSTMCells[i],
                    x if i == 0 else core.concat(h_list),
                    (h[i], c[i]),
                )
                hr_list, cr_list = self.cell_forward(
                    self.rLSTMCells[i],
                    x[::-1] if i == 0 else core.concat(hr_list),
                    (h[i + self.num_layers], c[i + self.num_layers]),
                )
                hn_list.append(h_list[-1])
                hrn_list.append(hr_list[-1])
                cn_list.append(c_list[-1])
                crn_list.append(cr_list[-1])
            output = core.concat(
                [core.concat(h_list),
                 core.concat(hr_list[::-1])], axis=-1)
            hn = core.concat(hn_list + hrn_list)
            cn = core.concat(cn_list + crn_list)
        if self.batch_first and x.ndim == 3:
            output = output.swapaxes(0, 1)
            hn = hn.swapaxes(0, 1)
            cn = cn.swapaxes(0, 1)

        return output, (hn, cn)

    def reset_parameters(self):
        for i in range(self.num_layers):
            self.LSTMCells[i].reset_paramters()
        if self.bidirectional:
            for i in range(self.num_layers):
                self.rLSTMCells[i].reset_paramters()

    def init_hidden(self, x):
        assert x.ndim in {2, 3}
        d = 2 if self.bidirectional else 1
        if x.ndim == 2:
            return zeros(
                (d * self.num_layers, self.hidden_size),
                **self.kwargs,
            )
        else:
            batch_size = x.shape[1]
            return zeros(
                (d * self.num_layers, batch_size, self.hidden_size),
                **self.kwargs,
            )

    def cell_forward(self, cell: RNNCell, x, h: Tuple):
        seq_len = x.shape[0]
        h_list, c_list = [], []
        for i in range(seq_len):
            h = cell(x[i], h)  # Infact, `h` here is a tuple (h, c)
            h_list.append(core.unsqueeze(h[0], axis=0))
            c_list.append(core.unsqueeze(h[1], axis=0))
        return h_list, c_list

    def __repr__(self) -> str:
        return "{}({}, {}, num_layers={}, bias={}, batch_first={}, bidirectional={})".format(
            self.__class__.__name__,
            self.input_size,
            self.hidden_size,
            self.num_layers,
            self.has_bias,
            self.batch_first,
            self.bidirectional,
        )

    def move(self, device):
        self.kwargs['device'] = device
        return super().move(device)


class GRUCell(Module):

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        bias: bool = True,
        device=None,
        dtype=None,
    ) -> None:
        super().__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.kwargs = {"device": Device(device), "dtype": dtype}

        self.Wx1 = Parameter(
            empty((input_size, 2 * hidden_size), **self.kwargs))
        self.Wh1 = Parameter(
            empty((hidden_size, 2 * hidden_size), **self.kwargs))
        self.Wx2 = Parameter(empty((input_size, hidden_size), **self.kwargs))
        self.Wh2 = Parameter(empty((hidden_size, hidden_size), **self.kwargs))

        if bias:
            self.bias1 = Parameter(empty(2 * self.hidden_size, **self.kwargs))
            self.bias2 = Parameter(empty(self.hidden_size, **self.kwargs))

        self.has_bias = bias
        self.reset_parameters()

    def forward(self, x, h=None):
        if h is None:
            h = self.init_hidden(x)
        else:
            assert (x.ndim == 1 and h.shape == (self.hidden_size, )) or (
                x.ndim == 2 and h.shape
                == (x.shape[0], self.hidden_size)), "Wrong hidden state input!"

        lin1 = x @ self.Wx1 + h @ self.Wh1
        if self.has_bias:
            lin1 = lin1 + self.bias1
        z, r = core.split(F.sigmoid(lin1), 2, axis=1)
        lin2 = x @ self.Wx2 + (r * h) @ self.Wh2
        if self.has_bias:
            lin2 = lin2 + self.bias2
        return (1 - z) * h + z * F.tanh(lin2)

    def reset_parameters(self):
        bound = math.sqrt(1 / self.hidden_size)
        init.uniform_(self.Wx1, -bound, bound)
        init.uniform_(self.Wx2, -bound, bound)
        init.uniform_(self.Wh1, -bound, bound)
        init.uniform_(self.Wh2, -bound, bound)
        if self.has_bias:
            init.uniform_(self.bias1, -bound, bound)
            init.uniform_(self.bias2, -bound, bound)

    def init_hidden(self, x):
        assert x.ndim in {1, 2}
        if x.ndim == 1:
            return zeros(self.hidden_size, **self.kwargs)
        else:
            batch_size = x.shape[0]
            return zeros((batch_size, self.hidden_size), **self.kwargs)

    def __repr__(self) -> str:
        return "{}({}, {}, bias={})".format(
            self.__class__.__name__,
            self.input_size,
            self.hidden_size,
            self.has_bias,
        )

    def move(self, device):
        self.kwargs['device'] = device
        return super().move(device)


class GRU(Module):

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_layers: int = 1,
        bias: bool = True,
        batch_first: bool = False,
        bidirectional: bool = False,
        device=None,
        dtype=None,
    ) -> None:
        super().__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.has_bias = bias
        self.batch_first = batch_first
        self.bidirectional = bidirectional
        self.kwargs = {"device": Device(device), "dtype": dtype}

        assert num_layers > 0
        size_list = [input_size] + [hidden_size] * (num_layers - 1)
        self.GRUCells: List[GRUCell] = []
        for i in range(num_layers):
            cell = GRUCell(
                size_list[i],
                hidden_size,
                bias,
                **self.kwargs,
            )
            setattr(self, 'gru_{}'.format(i), cell)
            self.GRUCells.append(cell)
        if self.bidirectional:
            self.rGRUCells: List[GRUCell] = []
            for i in range(num_layers):
                cell = GRUCell(
                    size_list[i],
                    hidden_size,
                    bias,
                    **self.kwargs,
                )
                setattr(self, 'rgru_{}'.format(i), cell)
                self.rGRUCells.append(cell)

    def forward(self, x, h=None):
        if self.batch_first and x.ndim == 3:
            x = x.swapaxes(0, 1)

        if h is None:
            h = self.init_hidden(x)
        else:
            d = 2 if self.bidirectional else 1
            assert (x.ndim == 2
                    and h.shape == (d * self.num_layers, self.hidden_size)
                    ) or (x.ndim == 3 and h.shape
                          == (d * self.num_layers, x.shape[1],
                              self.hidden_size)), "Wrong hidden state input!"

        if self.num_layers == 1 and not self.bidirectional:
            h_list = self.cell_forward(self.GRUCells[0], x, h[0])
            output = core.concat(h_list)
            hn = h_list[-1]

        elif self.num_layers == 1 and self.bidirectional:
            h_list = self.cell_forward(self.GRUCells[0], x, h[0])
            hr_list = self.cell_forward(self.rGRUCells[0], x[::-1], h[1])
            output = core.concat(
                [core.concat(h_list),
                 core.concat(hr_list[::-1])], axis=-1)
            hn = core.concat([h_list[-1], hr_list[-1]])

        elif self.num_layers > 1 and not self.bidirectional:
            hn_list = []
            for i in range(self.num_layers):
                h_list = self.cell_forward(
                    self.GRUCells[i],
                    x if i == 0 else core.concat(h_list),
                    h[i],
                )
                hn_list.append(h_list[-1])
            output = core.concat(h_list)
            hn = core.concat(hn_list)

        else:
            hn_list = []
            hrn_list = []
            for i in range(self.num_layers):
                h_list = self.cell_forward(
                    self.GRUCells[i],
                    x if i == 0 else core.concat(h_list),
                    h[i],
                )
                hr_list = self.cell_forward(
                    self.rGRUCells[i],
                    x[::-1] if i == 0 else core.concat(hr_list),
                    h[i + self.num_layers],
                )
                hn_list.append(h_list[-1])
                hrn_list.append(hr_list[-1])
            output = core.concat(
                [core.concat(h_list),
                 core.concat(hr_list[::-1])], axis=-1)
            hn = core.concat(hn_list + hrn_list)

        if self.batch_first and x.ndim == 3:
            output = output.swapaxes(0, 1)
            hn = hn.swapaxes(0, 1)
        return output, hn

    def init_hidden(self, x):
        assert x.ndim in {2, 3}
        d = 2 if self.bidirectional else 1
        if x.ndim == 2:
            return zeros(
                (d * self.num_layers, self.hidden_size),
                **self.kwargs,
            )
        else:
            return zeros(
                (d * self.num_layers, x.shape[1], self.hidden_size),
                **self.kwargs,
            )

    def cell_forward(self, cell: GRUCell, x, h):
        seq_len = x.shape[0]
        h_list = []
        for i in range(seq_len):
            h = cell(x[i], h)
            h_list.append(core.unsqueeze(h, axis=0))
        return h_list

    def __repr__(self) -> str:
        return "{}({}, {}, num_layers={}, bias={}, batch_first={}, bidirectional={})".format(
            self.__class__.__name__,
            self.input_size,
            self.hidden_size,
            self.num_layers,
            self.has_bias,
            self.batch_first,
            self.bidirectional,
        )

    def move(self, device):
        self.kwargs['device'] = device
        return super().move(device)



from .module import Module
from ..parameter import Parameter
from .. import init
from ...special import empty
from ... import core
from ...cuda import Device


class BatchNorm1d(Module):
    '''
    一维Batch Normalization层

    Parameters
    ----------
    num_features : int
        输入特征数.
    eps : float, default=1e-5
        防止除数为0的极小项.
    momentum : float, default=0.5
        计算累积均值和方差的动量项.
    device : Optional[Device], default=None
        层数据所在的设备.
    dtype : default=Nonr
        层数据的类型.
    '''

    def __init__(
        self,
        num_features: int,
        eps: float = 1e-6,
        momentum: float = 0.1,
        device=None,
        dtype=None,
    ) -> None:
        super().__init__()
        kwargs = {"device": Device(device), "dtype": dtype}
        self.num_features = num_features
        self.eps = eps
        self.momentum = momentum
        self.running_mean = Parameter(
            empty(self.num_features, **kwargs),
            requires_grad=False,
        )
        self.running_var = Parameter(
            empty(self.num_features, **kwargs),
            requires_grad=False,
        )
        self.scale = Parameter(empty(self.num_features, **kwargs))
        self.shift = Parameter(empty(self.num_features, **kwargs))
        self.reset_parameters()

    def reset_parameters(self):
        init.zeros_(self.running_mean)
        init.ones_(self.running_var)
        init.zeros_(self.shift)
        init.ones_(self.scale)

    def forward(self, x):
        if self._train:
            mean = x.mean(0)
            center_data = x - mean
            var = core.mean(core.square(center_data), 0)
            std_data = center_data / core.sqrt(var + self.eps)

            self.running_mean *= (1 - self.momentum)
            self.running_mean += self.momentum * mean
            self.running_var *= (1 - self.momentum)
            self.running_var += self.momentum * var

            return std_data * self.scale + self.shift
        else:
            return (x - self.running_mean) * self.scale / core.sqrt(
                self.running_var + self.eps) + self.shift

    def __repr__(self) -> str:
        return "{}(num_features={}, momentum={})".format(
            self.__class__.__name__,
            self.num_features,
            self.momentum,
        )


class BatchNorm2d(Module):
    '''
    二维Batch Normalization层

    Parameters
    ----------
    num_features : int
        输入特征数(通道数).
    eps : float, default=1e-5
        防止除数为0的极小项.
    momentum : float, default=0.5
        计算累积均值和方差的动量项.
    device : Optional[Device], default=None
        层数据所在的设备.
    dtype : default=None
        层数据的类型.
    '''

    def __init__(
        self,
        num_features: int,
        eps: float = 1e-6,
        momentum: float = 0.1,
        device=None,
        dtype=None,
    ) -> None:
        super().__init__()
        kwargs = {"device": Device(device), "dtype": dtype}
        self.num_features = num_features
        self.eps = eps
        self.momentum = momentum
        self.running_mean = Parameter(
            empty((1, self.num_features, 1, 1), **kwargs),
            requires_grad=False,
        )
        self.running_var = Parameter(
            empty((1, self.num_features, 1, 1), **kwargs),
            requires_grad=False,
        )
        self.scale = Parameter(empty(1, self.num_features, 1, 1, **kwargs))
        self.shift = Parameter(empty((1, self.num_features, 1, 1), **kwargs))
        self.reset_parameters()

    def reset_parameters(self):
        init.zeros_(self.running_mean)
        init.ones_(self.running_var)
        init.zeros_(self.shift)
        init.ones_(self.scale)

    def forward(self, x):
        if self._train:
            mean = x.mean((0, 2, 3), keepdims=True)
            center_data = x - mean
            var = core.mean(core.square(center_data), (0, 2, 3), keepdims=True)
            std_data = center_data / core.sqrt(var + self.eps)

            self.running_mean *= (1 - self.momentum)
            self.running_mean += self.momentum * mean
            self.running_var *= (1 - self.momentum)
            self.running_var += self.momentum * var

            return std_data * self.scale + self.shift
        else:
            return (x - self.running_mean) * self.scale / core.sqrt(
                self.running_var + self.eps) + self.shift

    def __repr__(self) -> str:
        return "{}(num_features={}, momentum={})".format(
            self.__class__.__name__,
            self.num_features,
            self.momentum,
        )


class LayerNorm(Module):
    '''
    Layer Normalization

    Parameters
    ----------
    normalized_shape : Tuple[int]
    eps : float, default=1e-5
    momentum : float, default=0.5
    device : Optional[Device], default=None
    dtype : default=None
    '''

    def __init__(
        self,
        normalized_shape: int,
        eps: float = 1e-6,
        momentum: float = 0.1,
        device=None,
        dtype=None,
    ) -> None:
        super().__init__()
        kwargs = {"device": Device(device), "dtype": dtype}
        if isinstance(normalized_shape, int):
            normalized_shape = (normalized_shape, )
        self.normalized_shape = tuple(normalized_shape)
        self.eps = eps
        self.momentum = momentum
        self.running_mean = Parameter(
            empty(normalized_shape, **kwargs),
            requires_grad=False,
        )
        self.running_var = Parameter(
            empty(normalized_shape, **kwargs),
            requires_grad=False,
        )
        self.scale = Parameter(empty(*normalized_shape, **kwargs))
        self.shift = Parameter(empty(normalized_shape, **kwargs))
        self.reset_parameters()

    def reset_parameters(self):
        init.zeros_(self.running_mean)
        init.ones_(self.running_var)
        init.zeros_(self.shift)
        init.ones_(self.scale)

    def forward(self, x):
        if self._train:
            axis = tuple(range(x.ndim - len(self.normalized_shape)))
            mean = x.mean(axis)
            center_data = x - mean
            var = core.square(center_data).mean(axis)
            std_data = center_data / core.sqrt(var + self.eps)
            self.running_mean *= (1 - self.momentum)
            self.running_mean += self.momentum * mean
            self.running_var *= (1 - self.momentum)
            self.running_var += self.momentum * var

            return std_data * self.scale + self.shift
        else:
            return (x - self.running_mean) * self.scale / core.sqrt(
                self.running_var + self.eps) + self.shift


class RMSNorm(Module):

    def __init__(
        self,
        normalized_shape: tuple,
        eps: float = 1e-6,
        device=None,
        dtype=None,
    ):
        super().__init__()
        kwargs = {"device": Device(device), "dtype": dtype}
        if isinstance(normalized_shape, int):
            normalized_shape = (normalized_shape, )
        self.normalized_shape = tuple(normalized_shape)
        self.sum_axis = tuple(
            [-(i + 1) for i in range(len(self.normalized_shape))])
        self.eps = eps

        self.weight = Parameter(empty(self.normalized_shape, **kwargs))
        self.reset_parameters()

    def reset_parameters(self):
        init.ones_(self.weight)

    def forward(self, x):
        z = core.square(x).mean(self.sum_axis, keepdims=True)
        z = x / core.sqrt(z + self.eps)
        return z * self.weight



from collections import OrderedDict

from ..parameter import Parameter
from ...core import Tensor
from ...autograd import set_grad_enabled
from ...cuda import Device, current_device


class Module:

    def __init__(self) -> None:
        self._train = True
        self.device = Device("cpu")
        self._parameters = OrderedDict()

    def __call__(self, *x) -> Tensor:
        return self.forward(*x)

    def __setattr__(self, __name: str, __value) -> None:
        self.__dict__[__name] = __value
        if isinstance(__value, Parameter):
            self._parameters[__name] = __value
        if isinstance(__value, Module):
            for key in __value._parameters:
                self._parameters[__name + "." + key] = __value._parameters[key]

    def __repr__(self) -> str:
        module_list = [
            module for module in self.__dict__.items()
            if isinstance(module[1], Module)
        ]
        return "{}(\n{}\n)".format(
            self.__class__.__name__,
            "\n".join([
                "{:>10} : {}".format(module_name, module)
                for module_name, module in module_list
            ]),
        )

    def parameters(self):
        for param in self._parameters.values():
            if param.requires_grad:
                yield param

    def train(self, mode: bool = True):
        set_grad_enabled(mode)
        self.set_module_state(mode)

    def set_module_state(self, mode: bool):
        self._train = mode
        for module in self.__dict__.values():
            if isinstance(module, Module):
                module.set_module_state(mode)

    def forward(self, x: Tensor) -> Tensor:
        raise NotImplementedError

    def eval(self):
        return self.train(False)

    def to(self, device):
        if not isinstance(device, Device):
            device = Device(device)
        if self.device != device:
            self.move(device)
        return self

    def move(self, device):
        self.device = device
        for module in self.__dict__.values():
            if isinstance(module, Module):
                module.move(device)
            if isinstance(module, Parameter):
                module.to(device)

    def cuda(self):
        return self.to(current_device())

    def cpu(self):
        return self.to('cpu')


class Sequential(Module):

    def __init__(self, *args) -> None:
        super().__init__()
        self.module_list = []
        if len(args) == 1 and isinstance(args[0], OrderedDict):
            for name, module in args[0].items():
                self.__setattr__(name, module)
                self.module_list.append(module)
        else:
            for idx, module in enumerate(args):
                self.__setattr__(str(idx), module)
                self.module_list.append(module)

    def forward(self, x: Tensor) -> Tensor:
        for module in self.module_list:
            x = module(x)
        return x

    def __len__(self):
        return len(self.module_list)


class ModuleList(Module):

    def __init__(self, module_list: list) -> None:
        super().__init__()
        self.module_list = module_list

        for idx, module in enumerate(module_list):
            self.__setattr__(str(idx), module)

    def __getitem__(self, index):
        return self.module_list[index]

    def __len__(self):
        return len(self.module_list)

    def append(self, module):
        self.module_list.append(module)
        self.__setattr__(str(len(self.module_list) - 1), module)

    def index(self, module):
        return self.module_list.index(module)



from ..core import Tensor


class Parameter(Tensor):

    def __init__(self, data: Tensor, requires_grad: bool = True) -> None:
        super().__init__(
            data=data.data,
            dtype=data.dtype,
            device=data.device,
            copy=False,
            requires_grad=requires_grad,
        )

    def __repr__(self) -> str:
        return "Parameter : \n{}".format(self.data) + (",\ndevice={}".format(
            self.device) if self.device.device != "cpu" else "")



import random
import sys
from pathlib import Path

import numpy as np
import pytest

sys.path.append(str(Path(__file__).resolve().parents[1]))

import pydynet as pdn

np.random.seed(0)
random.seed(0)


def _assert_allclose(actual, expected, atol=1e-6, rtol=1e-6):
    assert np.allclose(actual, expected, atol=atol, rtol=rtol)


def test_backward_scalar_polynomial():
    x = pdn.Tensor(2.0, requires_grad=True)
    y = x**2 + 3 * x - 1
    y.backward()
    _assert_allclose(x.grad, np.array(7.0))


def test_backward_broadcast_add():
    x_np = np.random.randn(2, 3).astype(np.float64)
    b_np = np.random.randn(1, 3).astype(np.float64)

    x = pdn.Tensor(x_np, requires_grad=True)
    b = pdn.Tensor(b_np, requires_grad=True)

    y = (x + b).sum()
    y.backward()

    _assert_allclose(x.grad, np.ones_like(x_np))
    _assert_allclose(b.grad, np.full_like(b_np, x_np.shape[0]))


def test_backward_matmul_sum():
    x_np = np.random.randn(2, 3).astype(np.float64)
    w_np = np.random.randn(3, 4).astype(np.float64)

    x = pdn.Tensor(x_np, requires_grad=True)
    w = pdn.Tensor(w_np, requires_grad=True)

    y = pdn.matmul(x, w).sum()
    y.backward()

    expected_x_grad = np.ones((2, 4), dtype=np.float64) @ w_np.T
    expected_w_grad = x_np.T @ np.ones((2, 4), dtype=np.float64)

    _assert_allclose(x.grad, expected_x_grad)
    _assert_allclose(w.grad, expected_w_grad)


def test_backward_retain_graph_twice_accumulates_grad():
    x = pdn.Tensor(2.0, requires_grad=True)
    y = x * x

    y.backward(retain_graph=True)
    first_grad = np.array(x.grad, copy=True)
    y.backward()

    _assert_allclose(first_grad, np.array(4.0))
    _assert_allclose(x.grad, np.array(8.0))


def test_backward_on_non_scalar_raises():
    x = pdn.Tensor(np.array([1.0, 2.0]), requires_grad=True)
    with pytest.raises(ValueError, match="scalar"):
        x.backward()
